//----------------------------------------------------------------------
//
//  FILE:   lpjio.cpp
//  AUTHOR: I. Ross
//  DATE:   19-MAY-2008
//
//----------------------------------------------------------------------
//
//  I/O file for standalone LPJ-WHyMe build using monthly data of
//  temperature, precipitation, cloudiness and number of wetdays from
//  the Climate Research Unit (CRU). Input and output is in netCDF
//  format.
//
//----------------------------------------------------------------------

// Copyright (C) 2010 Rita Wania <rita@wania.net>
//  
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//  
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//  
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//  

// There are two options that can be set in this driver file:
// 1. The model can be run only for peatland grid cells
//    Look for 2 sections that say PEATLAND GRID CELLS ONLY and follow 
//    instructions there.
//
// 2. The model can be run using either a wetdays climatology or timeseries
//    data. Look for all lines that say TIMESERIES and follow instructions.

//======================================================================
//
//  HEADER FILES
//
//======================================================================

#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <string>
#include <cstring>
#include <set>
#include <vector>
#include <cmath>

using namespace std;

#include <netcdf.hh>
#include "Array.hh"


//======================================================================
//
//  CONSTANT DEFINITIONS
//
//======================================================================

// Array sizes etc.

const int NMON = 12, NDAYS = 365, NTRACE = 6, NLAYER = 11;
const int NPFT = 12;


// Missing values to use for output files.

const float FMISSING = 1.0E35;
const int IMISSING = -999;

float maxtemp;

// If annual output is required, but it is not necessary to output
// all of the run years, some of the last few years can be output by
// setting the RUN_OUT_YEARS variable.

//const int run_out_years = 20;
//bool run_out_years_on = false;

//======================================================================
//
//  TYPE DEFINITIONS
//
//======================================================================

//----------------------------------------------------------------------
//
//  Run parameters: read from configuration file.
//

struct RunParameters {

  RunParameters() :
    in_data_dir("."),
    temp_var("tas"), dtemp_var("tas"), prec_var("pr"), dprec_var("pr"),
    sun_var("clt"), par_var("par"), dpar_var("par"), wetdays_var("wetdays"),
    mask_var("mask"), sun_is_cloud(true), mask_type(BOOLEAN),
    soil_type_fixed(true), fixed_soil_type(2), soil_type_var("soiltype"),
    spinup_years(1000), spinup_data_years(0), run_years(0),
    spinup_out_freq(0), run_out_freq(0), run_out_years(0), 
    run_out_years_on(true),
    perturb_tmean(0.0), perturb_trange(0.0), perturb_pmean(0.0)
  { }

  enum MET_DATA_TYPE { MONTHLY_MET = 0, PAR_DATA_MET = 1, DAILY_MET = 2 };
  enum MASK_TYPE { BOOLEAN, PERCENT_LAND };


  // Driver data directory.

  string in_data_dir;


  // Meterology data type.

  MET_DATA_TYPE met_data_type;

    
  // Driver data filenames and variable names (variable names default
  // to PMIP2 values).  The sun_is_cloud flag is used to indicate that
  // the input sun values are actually cloud.

  string temp_file, prec_file, sun_file, wetdays_file, mask_file;
  string dtemp_file, dprec_file, par_file, dpar_file;
  string temp_var, dtemp_var, prec_var, dprec_var, sun_var, par_var, dpar_var;
  string wetdays_var, mask_var;
  bool sun_is_cloud;
  MASK_TYPE mask_type;


  // Soil type data - either a single fixed value or a filename for a
  // file on the same grid as the climate data.

  bool soil_type_fixed;
  int fixed_soil_type;
  string soil_type_file;
  string soil_type_var;


  // Number of years of spin-up to do (default to 1000), the number of
  // years of the climate data to use for the spinup (zero means use
  // all years), and the number of years of transient run to do (zero
  // means use all the available data once).

  int spinup_years, spinup_data_years, run_years;


  // The interval in years between output records written during
  // spin-up (default is no output during spin-up, indicated by a zero
  // value) and the interval in years between output records written
  // during the main part of the run (output values are averaged over
  // these intervals, and the default, indicated by a zero value, is
  // to write a single averaged record at the end of the run).

  int spinup_out_freq, run_out_freq;


  int run_out_years;
  bool run_out_years_on;

  // CO2 concentration - either a time series, or a single value.

  vector<float> co2_conc;


  // Output filename.

  string output_file;


  // Output variables.

  struct OutVariable {
    OutVariable() :
      name(""), mean(false), sdev(false), max(false), min(false) { }
    OutVariable(string n) :
      name(n), mean(true), sdev(false), max(false), min(false) { }
    string name;
    bool mean, sdev, max, min;
  };

  vector<OutVariable> output_vars;

  // Climate perturbations.

  float perturb_tmean, perturb_trange;
  float perturb_pmean;
};


//----------------------------------------------------------------------
//
//  Variable grid definition: derived from driver files.
//

struct Grid {
  int nlat, nlon;
  vector<float> lats, lons;

  Grid() : nlat(0), nlon(0) { }
  Grid(string file);
  Grid(const Grid &other);
  Grid &operator=(const Grid &other);

  bool operator==(const Grid &other);
  bool operator!=(const Grid &other) { return !(*this == other); }
};


//----------------------------------------------------------------------
//
//  Types for handling LPJ output variables.
//

enum LPJ_VAR_TYPE {
  SIMPLE,
  SIMPLE_MONTHLY,
  SIMPLE_DAILY,
  PER_PFT,
  PER_PFT_MONTHLY,
  PER_PFT_DAILY,
  PER_LAYER_DAILY,
  TRACE,
  TRACE_MONTHLY,
};

struct LPJVariable {
  const char *name;
  LPJ_VAR_TYPE type;
};

struct LPJOutputVariable {
  LPJOutputVariable(LPJVariable *in_def, RunParameters::OutVariable &collect);
  ~LPJOutputVariable();
  void set_nc(int idx, NcFile &nc);
  void reset(void) { clear_buffs(); out_index = 0; }
  void clear_buffs(void);
  bool average_and_output(void);

  void transpose_buffers(int time_axis_len, int other_axis_len);

  LPJVariable *def;
  NcVar **mean_nc, **sdev_nc, **min_nc, **max_nc;
  int buff_size;
  float *avg_buff, *sdev_buff, *min_buff, *max_buff;
  bool mean, sdev, min, max;
  int accum_years;
  int out_index;
};


//======================================================================
//
//  LPJ VARIABLE DEFINITIONS
//
//======================================================================

// All the variable names we recognise from LPJ.

LPJVariable LPJ_VARIABLES[] = {
  { "nind",          PER_PFT         },
  { "lm_ind",        PER_PFT         },
  { "rm_ind",        PER_PFT         },
  { "sm_ind",        PER_PFT         },
  { "hm_ind",        PER_PFT         },
  { "vegc",          SIMPLE          },
  { "fpc_grid",      PER_PFT         },
  { "anpp",          PER_PFT         },
  { "acflux_estab",  SIMPLE          },
  { "litter_ag",     PER_PFT         },
  { "litter_bg",     PER_PFT         },
  { "cpool_fast",    SIMPLE          },
  { "cpool_slow",    SIMPLE          },
  { "arh",           SIMPLE          },
  { "aaresp",        SIMPLE          },
  { "aalresp",       SIMPLE          },
  { "alresp",        PER_PFT         },
  { "xalresp",       PER_PFT         },
  { "afire_frac",    SIMPLE          },
  { "acflux_fire",   SIMPLE          },
  { "arunoff",       SIMPLE          },
  { "sla",           PER_PFT         },
  { "mpar",          SIMPLE_MONTHLY  },
  { "mapar",         SIMPLE_MONTHLY  },
  { "anpp_add",      SIMPLE          },
  { "mrunoff",       SIMPLE_MONTHLY  },
  { "aaet",          SIMPLE          },
  { "mrh",           SIMPLE_MONTHLY  },
  { "mnpp_add",      SIMPLE_MONTHLY  },
  { "maet",          SIMPLE_MONTHLY  },
  { "mtemp_soil",    SIMPLE_MONTHLY  },
  { "mw1",           SIMPLE_MONTHLY  },
  { "mw2",           SIMPLE_MONTHLY  },
  { "maep",          SIMPLE_MONTHLY  },
  { "aaep",          SIMPLE          },
  { "maept",          SIMPLE_MONTHLY  },
  { "apet",          SIMPLE          },
  { "mintc",         SIMPLE_MONTHLY  },
  { "aintc",         SIMPLE          },
  { "gdd",           PER_PFT         },
  { "height",        PER_PFT         },
  { "mgpp",          PER_PFT_MONTHLY },
  { "mwtp",          SIMPLE_MONTHLY  },
  { "mthaw_depth",   SIMPLE_MONTHLY  },
  { "permafrost_index", SIMPLE       },
  { "msnowdepth",    SIMPLE_MONTHLY  },
  { "snowdepth",     SIMPLE_DAILY    },
  { "wtp",           SIMPLE_DAILY    },
  { "thawdepth",     SIMPLE_DAILY    },
  { "mnpp",          PER_PFT_MONTHLY },
  { "mtsoil_0",      SIMPLE_MONTHLY  },
  { "mtsoil_10",     SIMPLE_MONTHLY  },
  { "mtsoil_25",     SIMPLE_MONTHLY  },
  { "mtsoil_50",     SIMPLE_MONTHLY  },
  { "mtsoil_100",    SIMPLE_MONTHLY  },
  { "mtsoil_150",    SIMPLE_MONTHLY  },
  { "mtsoil_200",    SIMPLE_MONTHLY  },
  { "mice_25",       SIMPLE_MONTHLY  },
  { "mice_50",       SIMPLE_MONTHLY  },
  { "mice_100",      SIMPLE_MONTHLY  },
  { "mice_200",      SIMPLE_MONTHLY  },
  { "mprec_melt",    SIMPLE_MONTHLY  },
  { "CH4",           PER_LAYER_DAILY },
  { "CH4_diss",      PER_LAYER_DAILY },
  { "CH4_diss_mol",  PER_LAYER_DAILY },
  { "CH4_gas",       PER_LAYER_DAILY },
  { "CO2_soil",      PER_LAYER_DAILY },
  { "O2",            PER_LAYER_DAILY },
  { "CH4_diff",      SIMPLE_DAILY    },
  { "CH4_plant",     SIMPLE_DAILY    },
  { "CH4_plant_out", PER_LAYER_DAILY },
  { "CH4_ebull",     SIMPLE_DAILY    },
  { "CO2_diff",      SIMPLE_DAILY    },
  { "CO2_plant",     SIMPLE_DAILY    },
  { "O2_diff",       SIMPLE_DAILY    },
  { "CH4_oxid",      PER_LAYER_DAILY },
  { "O2_plant",      PER_LAYER_DAILY },
  { "CH4_prod",      PER_LAYER_DAILY },
  { "CO2_prod",      PER_LAYER_DAILY },
  { "Total_water",   PER_LAYER_DAILY },
  { "Tsoil",         PER_LAYER_DAILY },
  { "Fwater",        PER_LAYER_DAILY },
  { "Fice",          PER_LAYER_DAILY },
  { "drh",           SIMPLE_DAILY    },
  { "mCH4_plant",    SIMPLE_MONTHLY  },
  { "mCH4_diff",     SIMPLE_MONTHLY  },
  { "mCH4_ebull",    SIMPLE_MONTHLY  },
  { "wtot",          PER_LAYER_DAILY },
  { "drunoff_surf",  SIMPLE_DAILY    },
  { "drunoff_drain", SIMPLE_DAILY    },
  { "mcan_cond_tot", SIMPLE_MONTHLY  },
  { "mlai",          SIMPLE_MONTHLY  },
  { "mFwater1",      SIMPLE_MONTHLY  },
  { "mFice1",        SIMPLE_MONTHLY  },
  { "atsoil_25",     SIMPLE          },
  { "dphen",         PER_PFT_DAILY   },
  { "fbare",         SIMPLE          }

};


//======================================================================
//
//  GLOBAL VARIABLE DEFINITIONS
//
//======================================================================

RunParameters params;           // Run parameters.

Grid grid;                      // Common variable grid definition.

NcFile *clim_temp_nc;           // Climate data files.
NcFile *clim_dtemp_nc;
NcFile *clim_precip_nc;
NcFile *clim_dprecip_nc;
NcFile *clim_sun_nc;
NcFile *clim_dpar_nc;
NcFile *clim_par_nc;
NcFile *clim_wetdays_nc;

NcVar *clim_temp_var;           // Climate data variables.
NcVar *clim_dtemp_var;           // Climate data variables.
NcVar *clim_precip_var;
NcVar *clim_dprecip_var;
NcVar *clim_sun_var;
NcVar *clim_dpar_var;
NcVar *clim_par_var;
NcVar *clim_wetdays_var;

BoolArray *land_mask;           // Land mask.

IntArray *soil_type;            // Soil type data.

int no_of_months;               // Climate data size (in months).
int no_of_days;                 // Climate data size (in days).

float *clim_temp;               // Climate data for current point.
float *clim_dtemp;
//double *clim_dtemp;
float *clim_precip;
float *clim_dprecip;
//double *clim_dprecip;
float *clim_sun;
float *clim_par;
float *clim_dpar;
//double *clim_dpar;
float *clim_wetdays;

int land_points, points_done;   // Point counters.
int latidx, lonidx;

NcFile **out_nc;                // Output files.
int out_year_count;
int out_idx;
int *out_years;

NcDim *lat_dim;                 // Output dimensions.
NcDim *lon_dim;
NcDim *pft_dim;
NcDim *month_dim;               // Month index.
bool month_dim_needed = false;
NcDim *day_dim;                 // Day-of-year index.
bool day_dim_needed = false;
NcDim *trace_dim;               // Trace gas index.
bool trace_dim_needed = false;
NcDim *layer_dim;               // Soil layer index.
bool layer_dim_needed = false;

LPJOutputVariable **out_vars = 0; // Output variables.
int out_var_count = 0;


//======================================================================
//
//  LOCAL FUNCTION PROTOTYPES
//
//======================================================================

static void read_config(RunParameters &params);
static bool boolean_string(string value, int line, bool &error_flag);
static void read_mask(BoolArray *mask, string mask_file, Grid &grid,
                      string var, RunParameters::MASK_TYPE mask_type);
static void read_soil_type(IntArray *soil_type, string soil_type_file,
                           Grid &grid, string soil_type_var);
static void read_co2_file(string fname);
static void check_var_grid(NcVar *var, NcFile *nc,
                           string var_name, string file_name);
static void open_output_files(string file_name);
static void handle_output_record(string name, float *data);


//======================================================================
//
//  LPJ INTERFACE FUNCTION DEFINITIONS
//
//======================================================================

//--------------------------------------------------------------------
//
//  initio
//
//  Reads input climate and land mask files, initialises for looping
//  over the grid cells and time.
//   

extern "C" void initio_(int *met_data_type)
{
  // Read the configuration file.

  read_config(params);
  *met_data_type = params.met_data_type;


  // Extract grid information from input files.

  int ngrids = 1;
  if (params.temp_file != "")    ++ngrids;
  if (params.dtemp_file != "")   ++ngrids;
  if (params.prec_file != "")    ++ngrids;
  if (params.dprec_file != "")   ++ngrids;
  if (params.sun_file != "")     ++ngrids;
  if (params.par_file != "")     ++ngrids;
  if (params.dpar_file != "")    ++ngrids;
  if (params.wetdays_file != "") ++ngrids;
  Grid grids[ngrids];
  int igrid = 0;
  grids[igrid++] = Grid(params.mask_file);
  if (params.temp_file != "")    grids[igrid++] = Grid(params.temp_file);
  if (params.dtemp_file != "")   grids[igrid++] = Grid(params.dtemp_file);
  if (params.prec_file != "")    grids[igrid++] = Grid(params.prec_file);
  if (params.dprec_file != "")   grids[igrid++] = Grid(params.dprec_file);
  if (params.sun_file != "")     grids[igrid++] = Grid(params.sun_file);
  if (params.par_file != "")     grids[igrid++] = Grid(params.par_file);
  if (params.dpar_file != "")    grids[igrid++] = Grid(params.dpar_file);
  if (params.wetdays_file != "") grids[igrid++] = Grid(params.wetdays_file);
  for (int igrid = 1; igrid < ngrids; ++igrid)
    if (grids[igrid] != grids[0]) {
      cout << "igrid:" << igrid << endl;
    cout << "ERROR1: Incompatible grid definitions in driver files" << endl;
    exit(1);
  }
  if (!params.soil_type_fixed) {
    Grid soil_type_grid(params.soil_type_file);
    if (soil_type_grid != grids[0]) {
      cout << "ERROR2: Incompatible grid definitions in driver files" << endl;
      exit(1);
    }
  }
  grid = grids[0];


  // Read the land/sea mask and convert to a boolean array.

  land_mask = new BoolArray(grid.nlon, grid.nlat);
  read_mask(land_mask, params.mask_file, grid,
            params.mask_var, params.mask_type);


  // Read the soil type data, if necessary.

  if (!params.soil_type_fixed) {
    soil_type = new IntArray(grid.nlon, grid.nlat);
    read_soil_type(soil_type, params.soil_type_file,
                   grid, params.soil_type_var);
  }


  // Open all climate input data files.

  if (params.temp_file != "") {
    clim_temp_nc = new NcFile(params.temp_file.c_str());
    if (!clim_temp_nc->is_valid()) {
      cout << "ERROR1: could not open climate file " << params.temp_file << endl;
      exit(1);
    }
  }
  if (params.dtemp_file != "") {
    clim_dtemp_nc = new NcFile(params.dtemp_file.c_str());
    if (!clim_dtemp_nc->is_valid()) {
      cout << "ERROR2: could not open climate file " << params.dtemp_file << endl;
      exit(1);
    }
  }
  if (params.prec_file != "") {
    if (params.prec_file == params.temp_file)
      clim_precip_nc = clim_temp_nc;
    else {
      clim_precip_nc = new NcFile(params.prec_file.c_str());
      if (!clim_precip_nc->is_valid()) {
        cout << "ERROR3: could not open climate file " << params.prec_file << endl;
        exit(1);
      }
    }
  }
  if (params.dprec_file != "") {
    if (params.dprec_file == params.dtemp_file)
      clim_dprecip_nc = clim_dtemp_nc;
    else {
      clim_dprecip_nc = new NcFile(params.dprec_file.c_str());
      if (!clim_dprecip_nc->is_valid()) {
        cout << "ERROR4: could not open climate file " << params.dprec_file << endl;
        exit(1);
      }
    }
  }
  if (params.sun_file != "") {
    if (params.sun_file == params.temp_file)
      clim_sun_nc = clim_temp_nc;
    else if (params.sun_file == params.prec_file)
      clim_sun_nc = clim_precip_nc;
    else {
      clim_sun_nc = new NcFile(params.sun_file.c_str());
      if (!clim_sun_nc->is_valid()) {
        cout << "ERROR5: could not open climate file " << params.sun_file << endl;
        exit(1);
      }
    }
  }
  if (params.par_file != "") {
    if (params.par_file == params.temp_file)
      clim_par_nc = clim_temp_nc;
    else if (params.par_file == params.prec_file)
      clim_par_nc = clim_precip_nc;
    else {
      clim_par_nc = new NcFile(params.par_file.c_str());
      if (!clim_par_nc->is_valid()) {
        cout << "ERROR6: could not open climate file " << params.par_file << endl;
        exit(1);
      }
    }
  }
  if (params.dpar_file != "") {
    if (params.dpar_file == params.dtemp_file)
      clim_dpar_nc = clim_dtemp_nc;
    else if (params.dpar_file == params.dprec_file)
      clim_dpar_nc = clim_dprecip_nc;
    else {
      clim_dpar_nc = new NcFile(params.dpar_file.c_str());
      if (!clim_dpar_nc->is_valid()) {
        cout << "ERROR7: could not open climate file " << params.dpar_file << endl;
        exit(1);
      }
    }
  }
  if (params.wetdays_file != "") {
    clim_wetdays_nc = new NcFile(params.wetdays_file.c_str());
    if (!clim_wetdays_nc->is_valid()) {
      cout << "ERROR8: could not open climate file "
           << params.wetdays_file << endl;
      exit(1);
    }
  }

  // Access all climate variables.

  if (clim_temp_nc) {
    clim_temp_var = clim_temp_nc->get_var(params.temp_var.c_str());
    if (!clim_temp_var) {
      cout << "ERROR: could not access climate variable '"
           << params.temp_var << "'" << endl;
      exit(1);
    }
    check_var_grid(clim_temp_var, clim_temp_nc,
                   params.temp_var, params.temp_file);
  }
  if (clim_dtemp_nc) {
    clim_dtemp_var = clim_dtemp_nc->get_var(params.dtemp_var.c_str());
    if (!clim_dtemp_var) {
      cout << "ERROR: could not access climate variable '"
           << params.dtemp_var << "'" << endl;
      exit(1);
    }
    check_var_grid(clim_dtemp_var, clim_dtemp_nc,
                   params.dtemp_var, params.dtemp_file);
  }
  if (clim_precip_nc) {
    clim_precip_var = clim_precip_nc->get_var(params.prec_var.c_str());
    if (!clim_precip_var) {
      cout << "ERROR: could not access climate variable '"
           << params.prec_var << "'" << endl;
      exit(1);
    }
    check_var_grid(clim_precip_var, clim_precip_nc,
                   params.prec_var, params.prec_file);
  }
  if (clim_dprecip_nc) {
    clim_dprecip_var = clim_dprecip_nc->get_var(params.dprec_var.c_str());
    if (!clim_dprecip_var) {
      cout << "ERROR: could not access climate variable '"
           << params.dprec_var << "'" << endl;
      exit(1);
    }
    check_var_grid(clim_dprecip_var, clim_dprecip_nc,
                   params.dprec_var, params.dprec_file);
  }
  if (clim_sun_nc) {
    clim_sun_var = clim_sun_nc->get_var(params.sun_var.c_str());
    if (!clim_sun_var) {
      cout << "ERROR: could not access climate variable '"
           << params.sun_var << "'" << endl;
      exit(1);
    }
    check_var_grid(clim_sun_var, clim_sun_nc,
                   params.sun_var, params.sun_file);
  }
  if (clim_par_nc) {
    clim_par_var = clim_par_nc->get_var(params.par_var.c_str());
    if (!clim_par_var) {
      cout << "ERROR: could not access climate variable '"
           << params.par_var << "'" << endl;
      exit(1);
    }
    check_var_grid(clim_par_var, clim_par_nc,
                   params.par_var, params.par_file);
  }
  if (clim_dpar_nc) {
    clim_dpar_var = clim_dpar_nc->get_var(params.dpar_var.c_str());
    if (!clim_dpar_var) {
      cout << "ERROR: could not access climate variable '"
           << params.dpar_var << "'" << endl;
      exit(1);
    }
    check_var_grid(clim_dpar_var, clim_dpar_nc,
                   params.dpar_var, params.dpar_file);
  }
  if (clim_wetdays_nc) {
    clim_wetdays_var = clim_wetdays_nc->get_var(params.wetdays_var.c_str());
    if (!clim_wetdays_var) {
      cout << "ERROR: could not access climate variable '"
           << params.wetdays_var << "'" << endl;
      exit(1);
    }
    check_var_grid(clim_wetdays_var, clim_wetdays_nc,
                   params.wetdays_var, params.wetdays_file);

  }

  // Check the time bounds on the climate variables.  There should be
  // a whole number of years of data, and there should be the same
  // amount of data for each variable.

  if (params.met_data_type == RunParameters::MONTHLY_MET ||
      params.met_data_type == RunParameters::PAR_DATA_MET) {
    NcDim *temp_time_dim = clim_temp_var->get_dim(0);
    NcDim *precip_time_dim = clim_precip_var->get_dim(0);
    NcDim *sun_time_dim = params.met_data_type == RunParameters::MONTHLY_MET ?
      clim_sun_var->get_dim(0) : clim_par_var->get_dim(0);
    NcDim *wetdays_time_dim = clim_wetdays_var->get_dim(0);
    if (temp_time_dim->size() != precip_time_dim->size() ||
        temp_time_dim->size() != sun_time_dim->size()) {
      cout << "ERROR: incompatible time dimensions in climate files" << endl;
      exit(1);
    }
    // If using TIMESERIES data for wetdays comment these lines out
        if (wetdays_time_dim->size() != NMON) {
     cout << "ERROR: bad time dimension in wet days file" << endl;
     exit(1);
    }
    no_of_months = temp_time_dim->size();
    if (no_of_months % NMON != 0) {
      cout << "ERROR: non whole number of years of climate data!" << endl;
      exit(1);
    }
    clim_temp = new float[no_of_months];
    clim_precip = new float[no_of_months];
    if (params.met_data_type == RunParameters::MONTHLY_MET)
      clim_sun = new float[no_of_months];
    else
      clim_par = new float[no_of_months];
    // If using TIMESERIES data for wetdays comment this line out
        clim_wetdays = new float[NMON];
    // If using TIMESERIES data for wetdays uncomment this line
    // clim_wetdays = new float[no_of_months];
  } else {
    NcDim *temp_time_dim = clim_temp_var->get_dim(0);
    // If using TIMESERIES data for wetdays uncomment this line
    // NcDim *wetdays_time_dim = clim_wetdays_var->get_dim(0); 
    NcDim *dtemp_time_dim = clim_dtemp_var->get_dim(0);
    NcDim *dprecip_time_dim = clim_dprecip_var->get_dim(0);
    NcDim *par_time_dim = clim_par_var->get_dim(0);
    //    NcDim *dpar_time_dim = clim_dpar_var->get_dim(0);
    NcDim *precip_time_dim = clim_precip_var->get_dim(0);
    if (dtemp_time_dim->size() != dprecip_time_dim->size() ||
        //        dtemp_time_dim->size() != dpar_time_dim->size() ||
        dtemp_time_dim->size() != temp_time_dim->size() / 12 * 365) {
      cout << "ERROR: incompatible time dimensions in climate files" << endl;
      exit(1);
    }
    no_of_months = temp_time_dim->size();
    no_of_days = dtemp_time_dim->size();
    if (no_of_months % NMON != 0 || no_of_days % NDAYS != 0) {
      cout << "ERROR: non whole number of years of climate data!" << endl;
      exit(1);
    }
    clim_temp = new float[no_of_months];
    clim_precip = new float[no_of_months];
    clim_par = new float[no_of_months];
    clim_dtemp = new float[no_of_days];
    clim_dprecip = new float[no_of_days];
    // If using TIMESERIES data for wetdays uncomment this line
    // clim_wetdays = new float[no_of_months]; 
    //clim_dtemp = new double[no_of_days];
    //clim_dprecip = new double[no_of_days];
    //clim_dpar = new float[no_of_days];
    //clim_dpar = new double[no_of_days];

  }


  // Count land points.

  land_points = 0;
  for (int lat = 0; lat < grid.nlat; ++lat)
    for (int lon = 0; lon < grid.nlon; ++lon)
      if ((*land_mask)(lon, lat)) ++land_points;


  // Determine number of output files.

  out_year_count = 0;
  if (params.spinup_out_freq != 0)
    out_year_count += params.spinup_years / params.spinup_out_freq;
  if (params.run_out_freq == 0)
    out_year_count += 1;
  else {
    if (params.run_years == 0)
      out_year_count += no_of_months / NMON / params.run_out_freq;
    else
      out_year_count += params.run_years / params.run_out_freq;
  }

  // C_RS: uncomment next line for the last 15 years output
  if (params.run_out_years_on)  out_year_count = params.run_out_years;


  out_years = new int[out_year_count];
  int out_year_idx = 0;
  if (params.spinup_out_freq != 0) {
    for (int idx = 1;
         idx <= params.spinup_years / params.spinup_out_freq; ++idx)
      out_years[out_year_idx++] = idx * params.spinup_out_freq;
  }
  if (params.run_out_freq == 0) {
    if (params.run_years == 0)
      out_years[out_year_idx] = params.spinup_years + no_of_months / NMON;
    else
      out_years[out_year_idx] = params.spinup_years + params.run_years;
  } else {
    for (int out_year = params.spinup_years + params.run_out_freq;
         out_year_idx < out_year_count; out_year += params.run_out_freq)
      out_years[out_year_idx++] = out_year;
  }
  // C_RS: uncomment next block for the last x (dependent on
  //  params.run_out_years) years output
  if (params.run_out_years_on){ 
    out_year_idx = 0;
    for (int out_year = params.spinup_years + params.run_years 
           - (out_year_count - 1);
         out_year_idx < out_year_count; out_year += 1)
      out_years[out_year_idx++] = out_year; 
  }

  // Diagnostic output.

  cout << "Total land points:  " << land_points << endl;
  cout << "Spinup years:       " << params.spinup_years << endl;
  cout << "Spinup output freq: ";
  if (params.spinup_out_freq != 0)
    cout << params.spinup_out_freq << endl;
  else
    cout << "n/a" << endl;
  cout << "Run years:          ";
  if (params.run_years != 0)
    cout << params.run_years << endl;
  else
    cout << "all available" << endl;
  cout << "Run output freq:    ";
  if (params.run_out_freq != 0)
    cout << params.run_out_freq << endl;
  else
    cout << "single output file" << endl;


  // Initialise the grid point counter.

  points_done = 0;


  // Find first valid point.

  // PEATLAND GRID CELLS ONLY
  // When running lpj only for peatland grid cells (soiltype = 8) use
  // these lines
     // latidx = lonidx = 0;
     // while (latidx < grid.nlat &&
     //       ((*land_mask)(lonidx, latidx) == false ||
     //        (*soil_type)(lonidx, latidx) != 8)) {
     //  ++lonidx;
     //  if (lonidx >= grid.nlon) { lonidx = 0; ++latidx; }
     // }
//  Otherwise use these
  latidx = lonidx = 0;
  while (latidx < grid.nlat && !(*land_mask)(lonidx, latidx)) {
    ++lonidx;
    if (lonidx >= grid.nlon) { lonidx = 0; ++latidx; }
  }

  // Open the output file (also initialises variable definitions and
  // buffers).

  open_output_files(params.output_file);
}


//----------------------------------------------------------------------
//
//  getgrid
//
//  Find the location of the next land point to process and pass the
//  latitude and soil type to the main LPJ module.  Also deal with
//  end condition if there are no more points to process.

extern "C" int getgrid_(float *latitude, int *soiltype, int *dogridcell)
{
  // Set up return values.
  if (latidx >= grid.nlat)
    *dogridcell = 0;
  else {
    *dogridcell = 1;
    if (params.soil_type_fixed)
      *soiltype = params.fixed_soil_type;
    else
      *soiltype = (*soil_type)(lonidx, latidx);
    *latitude = grid.lats[latidx];

    // Read climate data for this point (note that the windspeed data
    // we're using here is only a seasonal cycle, not a full monthly
    // time series like the other data).

    if (clim_temp)
      if (!clim_temp_var->set_cur(0, latidx, lonidx) ||
          !clim_temp_var->get(clim_temp, no_of_months, 1, 1)) {
        cout << "Failed to read temperature data" << endl;
        exit(1);
      }
    if (clim_dtemp)
      if (!clim_dtemp_var->set_cur(0, latidx, lonidx) ||
          !clim_dtemp_var->get(clim_dtemp, no_of_days, 1, 1)) {
        cout << "Failed to read temperature data" << endl;
        exit(1);
      }
    if (clim_precip)
      if (!clim_precip_var->set_cur(0, latidx, lonidx) ||
          !clim_precip_var->get(clim_precip, no_of_months, 1, 1)) {
        cout << "Failed to read precipitation data" << endl;
        exit(1);
      }
    if (clim_dprecip)
      if (!clim_dprecip_var->set_cur(0, latidx, lonidx) ||
          !clim_dprecip_var->get(clim_dprecip, no_of_days, 1, 1)) {
        cout << "Failed to read precipitation data" << endl;
        exit(1);
      }
    if (clim_sun)
      if (!clim_sun_var->set_cur(0, latidx, lonidx) ||
          !clim_sun_var->get(clim_sun, no_of_months, 1, 1)) {
        cout << "Failed to read cloudiness data" << endl;
        exit(1);
      }
    if (clim_par)
      if (!clim_par_var->set_cur(0, latidx, lonidx) ||
          !clim_par_var->get(clim_par, no_of_months, 1, 1)) {
        cout << "Failed to read cloudiness data" << endl;
        exit(1);
      }
    if (clim_dpar)
      if (!clim_dpar_var->set_cur(0, latidx, lonidx) ||
          !clim_dpar_var->get(clim_dpar, no_of_days, 1, 1)) {
        cout << "Failed to read cloudiness data" << endl;
        exit(1);
      }
    if (clim_wetdays)
      if (!clim_wetdays_var->set_cur(0, latidx, lonidx) ||
    // If using TIMESERIES data for wetdays comment this line out
          !clim_wetdays_var->get(clim_wetdays, NMON, 1, 1)) {
    // If using TIMESERIES data for wetdays uncomment this line
    //      !clim_wetdays_var->get(clim_wetdays, no_of_months, 1, 1)) {
        cout << "Failed to read wet days data" << endl;
        exit(1);
      }

    // Convert units: temperature should be in deg. C (input is K);
    // precipitation should be in mm/month (input is in kg m-2 s-1);
    // sunshine should be in percent (input is percentage cloud).

    for (int mon = 0; mon < no_of_months; ++mon) {
      // When using the T in K and P in kg m-2 s-1 uncomment the next 2 lines
      // The timeseries data are in deg C in mm per month already
      // For future scenarios, check which data you use!
      // When using the ECMWF ERA40 or ECHAM5 data uncomment the 'temp' line
      //            if (clim_temp) clim_temp[mon] -= 273.16;
      //            if (clim_precip) clim_precip[mon] *= 3600 * 24 * 30;

      if (clim_sun && params.sun_is_cloud) clim_sun[mon] = 100.0 - clim_sun[mon];
    }
    // for (int day = 0; day < no_of_days; ++day) {
    //   if (clim_dtemp) clim_dtemp[mon] -= 273.16;
    //   if (clim_dprecip) clim_dprecip[mon] *= 3600 * 24 * 30;
    // }

    // Perform climate perturbations if required.

    if (fabs(params.perturb_tmean) > 1.0E-6) {
      for (int mon = 0; mon < no_of_months; ++mon)
        clim_temp[mon] += params.perturb_tmean;
    }

//     // Increase the T range for each season. This is not an ideal
//     // solution for future climate change scenarios, as all months see
//     // a warming trend in general, but some months warm more than
//     // others, depending on the latitude.

//     if (fabs(params.perturb_trange) > 1.0E-6) {
//       // Find maximum and minimum temperature through climate data,
//       // along with temperature sum (for mean calculation).
//       float min_temp = clim_temp[0], max_temp = clim_temp[0];
//       float temp_sum = 0.0;
//       for (int mon = 0; mon < no_of_months; ++mon) {
//         temp_sum += clim_temp[mon];
//         if (clim_temp[mon] < min_temp) min_temp = clim_temp[mon];
//         if (clim_temp[mon] > max_temp) max_temp = clim_temp[mon];
//       }

//       // Find temperature mean and original range.
//       float temp_mean = temp_sum / no_of_months;
//       float temp_range = max_temp - min_temp;

//       // If we want to increase the minimum to maximum range of
//       // temperature by params.perturb_trange, we need to expand the
//       // range by this factor - it's just the new range (the old range
//       // plus the perturbation) divided by the old range.
//       float range_factor = (temp_range + params.perturb_trange) / temp_range;

//       // For each temperature entry, do the following:
//       //  - Take off the mean (to give an anomaly from the mean value);
//       //  - Multiply the anomaly by our range expansion factor;
//       //  - Put the mean back on.
//       for (int mon = 0; mon < no_of_months; ++mon)
//         clim_temp[mon] =
//           (clim_temp[mon] - temp_mean) * range_factor + temp_mean;

//       // The end result of all this is a temperature time series with
//       // the same mean as before, but with a minimum to maximum range
//       // expanded symmetrically about the mean by the amount we asked
//       // for.
//     }

    // Increase the T more in some months than in others. In low
    // latitudes (<60 deg N), the summer Ts increase much more than
    // the winter Ts, whereas >60deg N, the opposite pattern is
    // found. To reflect this pattern in the sensitivity studies, use
    // a sin function to increase the summer month Ts more and a cos
    // function to increase the winter Ts more.

    // maxtemp = the mean trange value multiplied by a factor you get
    // from summing over all months for fabs(sin(15 * mon * M_PI /
    // 180)). Then you get a mean value and divide the maximum value
    // (=1) by the mean value (~0.63). This gives ~1.58.

    maxtemp = params.perturb_trange * 1.5815793;
    if (fabs(params.perturb_trange) > 1.0E-6) {
      if (fabs(*latitude) < 60.0) {
      for (int mon = 0; mon < no_of_months; ++mon)
        clim_temp[mon] =
          clim_temp[mon] + maxtemp * fabs(sin(15 * mon * M_PI / 180));
      } else
      for (int mon = 0; mon < no_of_months; ++mon)
        clim_temp[mon] =
          clim_temp[mon] + maxtemp * fabs(cos(15 * mon * M_PI / 180));
    }

    if (fabs(params.perturb_pmean) > 1.0E-6) {
      for (int mon = 0; mon < no_of_months; ++mon)
        clim_precip[mon] *= (1.0 + params.perturb_pmean / 100.0);
    }


    // Reset all output variables.

    for (int idx = 0; idx < out_var_count; ++idx) out_vars[idx]->reset();
    out_idx = 0;
  }
  return 0;
}


//----------------------------------------------------------------------
//
//  getclimate
//
//  Sends this year's climatology to the main program.  Also deals
//  with end condition for time.

bool getclimate_common(int year, int &timeidx, float &co2, int &doyear)
{
  // The total length of the simulation is params.spinup_years of
  // spin-up, followed by as many years as we have data for.

  int run_years = params.run_years == 0 ? no_of_months / NMON : params.run_years;
  if (year > params.spinup_years + run_years) {
    doyear = 0;
    return false;
  } else
    doyear = 1;
  if (year > params.spinup_years)
    timeidx = year - params.spinup_years - 1;
  else
    timeidx = (year - 1) % (params.spinup_data_years ?
                            params.spinup_data_years : (no_of_months / NMON));
  if (timeidx >= no_of_months / NMON) timeidx %= no_of_months / NMON;


  // Set up the CO2 concentration -- this is either a constant value
  // (stored in a one element vector in params.co2_conc), or a time
  // series, from which we use the first value continuously for
  // spin-up, then successive values for the transient part of the
  // run.

  if (params.co2_conc.size() == 1)
    co2 = params.co2_conc[0];
  else if (year > params.spinup_years)
    co2 = params.co2_conc[year - params.spinup_years - 1];
  else
    co2 = params.co2_conc[0];

  return true;
}

extern "C" int getclimatemonthly_(int *year,
                                  float *mtemp, float *mprec, float *mwet,
                                  float *msunpar, float *co2, int *doyear,
                                  float *mtemp_prev, float *mtemp_next,
                                  int *type)
{
  int timeidx;
  if (!getclimate_common(*year, timeidx, *co2, *doyear)) return 0;
  timeidx *= NMON;


  // Copy climate data for the current month.

  for (int mon = 0; mon < NMON; ++mon) {
    mtemp[mon] = clim_temp[timeidx + mon];
    mprec[mon] = clim_precip[timeidx + mon];
    msunpar[mon] = *type == RunParameters::PAR_DATA_MET ?
      clim_par[timeidx + mon] : clim_sun[timeidx + mon];
    // If using TIMESERIES data for wetdays comment this line out
    mwet[mon] = clim_wetdays[mon];
    // If using TIMESERIES data for wetdays uncomment this line
    //mwet[mon] = clim_wetdays[timeidx + mon];
    if (timeidx + mon == 0)
      mtemp_prev[mon] = clim_temp[timeidx + mon];
    else
      mtemp_prev[mon] = clim_temp[timeidx - 1 + mon];
    if (timeidx + mon == no_of_months - 1)
      mtemp_next[mon] = clim_temp[timeidx + mon];
    else
      mtemp_next[mon] = clim_temp[timeidx + 1 + mon];
  }

  return 0;
}


extern "C" int getclimatedaily_(int *year,
                                float *dtemp, float *dprec, float *msunpar,
                                float *mtemp, float *mprec,
                                float *co2, int *doyear, int *type)
{
  int mtimeidx;
  if (!getclimate_common(*year, mtimeidx, *co2, *doyear)) return 0;
  int dtimeidx = mtimeidx * NDAYS;
  mtimeidx *= NMON;


  // Copy climate data for the current month.

  for (int mon = 0; mon < NMON; ++mon) {
    mtemp[mon] = clim_temp[mtimeidx + mon];
    mprec[mon] = clim_precip[mtimeidx + mon];
    msunpar[mon] = *type == RunParameters::DAILY_MET ?
      clim_par[mtimeidx + mon] : clim_sun[mtimeidx + mon];
  }
  for (int day = 0; day < NDAYS; ++day) {
    dtemp[day] = clim_dtemp[dtimeidx + day];
    dprec[day] = clim_dprecip[dtimeidx + day];
  }

  return 0;
}


//----------------------------------------------------------------------
//
//  outannual
//
//  Collect a year's worth of output for one grid cell.

extern "C" int outannual_(int *year, int *present,
                          float *nind,
                          float *lm_ind, float *rm_ind,
                          float *sm_ind, float *hm_ind,
                          float *vegc, float *fpc_grid,
                          float *anpp, float *acflux_estab,
                          float *litter_ag, float *litter_bg,
                          float *cpool_fast, float *cpool_slow,
                          float *arh, float *aaresp, float *aalresp,
                          float *alresp, float *xalresp,
                          float *afire_frac, float *acflux_fire,
                          float *arunoff, float *sla, float *mpar,
                          float *mapar,  float *anpp_add,
                          float *mrunoff, float *aaet,
                          float *mrh, float *mnpp_add, float *maet,
                          float *mtemp_soil, 
                          float *mw1, float *mw2,
                          float *maep, float *aaep, float *maept, float *apet, 
                          float *mintc, float *aintc,
                          float *gdd, float *height, float *mgpp,
                          float *mwtp, float *mthaw_depth,
			  float *permafrost_index, float *msnowdepth,
                          float *snowdepth,
                          float *wtp, float *thawdepth, float* mnpp,
                          float *mtsoil_0, float *mtsoil_10, float *mtsoil_25,
                          float *mtsoil_50, float *mtsoil_100,float *mtsoil_150,
                          float *mtsoil_200, float *mice_25, float *mice_50,
                          float *mice_100, float *mice_200, float *mprec_melt,
                          float *CH4, float *CH4_diss, float *CH4_diss_mol,
                          float *CH4_gas, 
                          float *CO2_soil, 
                          float *O2, float *CH4_diff,  float *CH4_plant,
                          float *CH4_plant_out,
                          float *CH4_ebull,  float *CO2_diff,  float *CO2_plant,
                          float *O2_diff, float *CH4_oxid, float *O2_plant,
                          float *CH4_prod, float *CO2_prod, float *Total_water, 
                          float *Tsoil, float *Fwater, float *Fice, float *drh,
                          float *mCH4_plant, float *mCH4_diff,
                          float *mCH4_ebull, float *wtot, float *drunoff_surf,
                          float *drunoff_drain, float *mcan_cond_tot, float *mlai,
                          float *mFwater1, float *mFice1, float* atsoil_25,
                          float *dphen
)
{
  // Skip output during spin-up if required.

  if (*year <= params.spinup_years && params.spinup_out_freq == 0)
    return 0;


  // Calculate bare ground fraction.

  float fbare = 0.0;
  for (int idx = 0; idx < NPFT; ++idx) fbare += fpc_grid[idx];
  fbare = 1.0 - fbare;
  if (fbare > 1.0) fbare = 1.0;
  if (fbare < 0.0) fbare = 0.0;


  // Accumulate values for averaging.

  handle_output_record("nind", nind);
  handle_output_record("lm_ind", lm_ind);
  handle_output_record("rm_ind", rm_ind);
  handle_output_record("sm_ind", sm_ind);
  handle_output_record("hm_ind", hm_ind);
  handle_output_record("vegc", vegc);
  handle_output_record("fpc_grid", fpc_grid);
  handle_output_record("anpp", anpp);
  handle_output_record("acflux_estab", acflux_estab);
  handle_output_record("litter_ag", litter_ag);
  handle_output_record("litter_bg", litter_bg);
  handle_output_record("cpool_fast", cpool_fast);
  handle_output_record("cpool_slow", cpool_slow);
  handle_output_record("arh", arh);
  handle_output_record("aaresp", aaresp);
  handle_output_record("aalresp", aalresp);
  handle_output_record("alresp", alresp);
  handle_output_record("xalresp", xalresp);
  handle_output_record("afire_frac", afire_frac);
  handle_output_record("acflux_fire", acflux_fire);
  handle_output_record("arunoff", arunoff);
  handle_output_record("sla", sla);
  handle_output_record("mpar", mpar);
  handle_output_record("mapar", mapar);
  handle_output_record("anpp_add", anpp_add);
  handle_output_record("mrunoff", mrunoff);
  handle_output_record("aaet", aaet);
  handle_output_record("mrh", mrh);
  handle_output_record("mnpp_add", mnpp_add);
  handle_output_record("maet", maet);
  handle_output_record("mtemp_soil", mtemp_soil);
  handle_output_record("mw1", mw1);
  handle_output_record("mw2", mw2);
  handle_output_record("maep", maep);
  handle_output_record("aaep", aaep);
  handle_output_record("maept", maept);
  handle_output_record("apet", apet);
  handle_output_record("mintc", mintc);
  handle_output_record("aintc", aintc);
  handle_output_record("gdd", gdd);
  handle_output_record("height", height);
  handle_output_record("mgpp", mgpp);
  handle_output_record("mwtp", mwtp);
  handle_output_record("mthaw_depth", mthaw_depth);
  handle_output_record("permafrost_index", permafrost_index);
  handle_output_record("msnowdepth", msnowdepth);
  handle_output_record("snowdepth", snowdepth);
  handle_output_record("wtp", wtp);
  handle_output_record("thawdepth", thawdepth);
  handle_output_record("mnpp", mnpp);
  handle_output_record("mtsoil_0", mtsoil_0);
  handle_output_record("mtsoil_10", mtsoil_10);
  handle_output_record("mtsoil_25", mtsoil_25);
  handle_output_record("mtsoil_50", mtsoil_50);
  handle_output_record("mtsoil_100", mtsoil_100);
  handle_output_record("mtsoil_150", mtsoil_150);
  handle_output_record("mtsoil_200", mtsoil_200);
  handle_output_record("mice_25", mice_25);
  handle_output_record("mice_50", mice_50);
  handle_output_record("mice_100", mice_100);
  handle_output_record("mice_200", mice_200);
  handle_output_record("mprec_melt", mprec_melt);
  handle_output_record("CH4", CH4);
  handle_output_record("CH4_diss", CH4_diss);
  handle_output_record("CH4_diss_mol", CH4_diss_mol);
  handle_output_record("CH4_gas", CH4_gas);
  handle_output_record("CO2_soil", CO2_soil);
  handle_output_record("O2", O2);
  handle_output_record("CH4_diff", CH4_diff);
  handle_output_record("CH4_plant", CH4_plant);
  handle_output_record("CH4_plant_out", CH4_plant_out);
  handle_output_record("CH4_ebull", CH4_ebull);
  handle_output_record("CO2_diff", CO2_diff);
  handle_output_record("CO2_plant", CO2_plant);
  handle_output_record("O2_diff", O2_diff);
  handle_output_record("CH4_oxid", CH4_oxid);
  handle_output_record("O2_plant", O2_plant);
  handle_output_record("CH4_prod", CH4_prod);
  handle_output_record("CO2_prod", CO2_prod);
  handle_output_record("Total_water", Total_water);
  handle_output_record("Tsoil", Tsoil);
  handle_output_record("Fwater", Fwater);
  handle_output_record("Fice", Fice);
  handle_output_record("drh", drh);
  handle_output_record("mCH4_plant", mCH4_plant);
  handle_output_record("mCH4_diff", mCH4_diff);
  handle_output_record("mCH4_ebull", mCH4_ebull);
  handle_output_record("wtot", wtot);
  handle_output_record("drunoff_surf", drunoff_surf);
  handle_output_record("drunoff_drain", drunoff_drain);
  handle_output_record("mcan_cond_tot", mcan_cond_tot);
  handle_output_record("mlai", mlai);
  handle_output_record("mFwater1", mFwater1);
  handle_output_record("mFice1", mFice1);
  handle_output_record("atsoil_25", atsoil_25);
  handle_output_record("dphen", dphen);
  handle_output_record("fbare", &fbare);


  // Output is written every params.spinup_out_freq years during
  // spin-up (averaged over that period).  During the transient part
  // of the run, data is either collected to write an average value
  // for each relevant variable at the end of the run, or average
  // values are written out every params.run_out_freq years.  A zero
  // value for params.spinup_out_freq means to write no output during
  // spin-up.

  if (*year <= params.spinup_years) {
    if (*year % params.spinup_out_freq != 0)
      return 0;
  } else {
    int run_years =
      params.run_years == 0 ? no_of_months / NMON : params.run_years;
    if (params.run_out_freq == 0) {
      if (*year < params.spinup_years + run_years)
        return 0;
    } else {
      // C_RS: uncomment line below for the last 15 years
        if (params.run_out_years_on){
          if ((params.spinup_years + run_years - *year) >= out_year_count)
            return 0;
        }  else {
          if ((*year - params.spinup_years) % params.run_out_freq != 0)
            return 0;
        }
    }
  }


  // Average and output variables and reinitialise averaging buffers.

  for (int idx = 0; idx < out_var_count; ++idx)
    if (!out_vars[idx]->average_and_output()) {
      cout << "ERROR: failed writing variable: "
           << out_vars[idx]->def->name << endl;
      exit(1);
    }
  out_nc[out_idx++]->sync();

  return 0;
}


//----------------------------------------------------------------------
//
//  outgrid
//
//  Write progress message and find next valid point.
//

extern "C" int outgrid_(void)
{
  // Progress message.

  if (++points_done % 1000 == 0)
    cout << points_done << " points processed ("
         << points_done * 100 / land_points << "%)" << endl;


  // Find next valid point.

  // PEATLAND GRID CELLS ONLY
  // When running lpj only for peatland grid cells (soiltype = 8) use
  // these lines
     //  ++lonidx;
     // if (lonidx >= grid.nlon) { lonidx = 0; ++latidx; }
     // while (latidx < grid.nlat &&
     //       ((*land_mask)(lonidx, latidx) == false ||
     //        (*soil_type)(lonidx, latidx) != 8)) {
     //  ++lonidx;
     //  if (lonidx >= grid.nlon) { lonidx = 0; ++latidx; }
     // }
  //  Otherwise use these
  ++lonidx;
  if (lonidx >= grid.nlon) { lonidx = 0; ++latidx; }
  while (latidx < grid.nlat && !(*land_mask)(lonidx, latidx)) {
    ++lonidx;
    if (lonidx >= grid.nlon) { lonidx = 0; ++latidx; }
  }

  return 0;
}


//----------------------------------------------------------------------
//
//  termio
//
//  Close output file.
//

extern "C" int termio_(void)
{
  for (int idx = 0; idx < out_year_count; ++idx) delete out_nc[idx];
  delete [] out_nc;
  return 0;
}


//======================================================================
//
//  LOCAL FUNCTION DEFINITIONS
//
//======================================================================

//----------------------------------------------------------------------
//
//  LPJOutputVariable::LPJOutputVariable
//
//  Constructor to create a single variable entry - contains
//  information about the variable definition, the output netCDF
//  variable, and an averaging buffer.
//

LPJOutputVariable::LPJOutputVariable(LPJVariable *in_def,
                                     RunParameters::OutVariable &collect)
{
  def = in_def;
  switch (def->type) {
  case SIMPLE:          buff_size = 1;             break;
  case SIMPLE_MONTHLY:  buff_size = NMON;          break;
  case SIMPLE_DAILY:    buff_size = NDAYS;         break;
  case PER_PFT:         buff_size = NPFT;          break;
  case PER_PFT_MONTHLY: buff_size = NMON * NPFT;   break;
  case PER_PFT_DAILY:   buff_size = NDAYS * NPFT;  break;
  case PER_LAYER_DAILY: buff_size = NDAYS * NLAYER;break;
  case TRACE:           buff_size = NTRACE;        break;
  case TRACE_MONTHLY:   buff_size = NTRACE * NMON; break;
  }
  if (def->type == SIMPLE_MONTHLY ||
      def->type == PER_PFT_MONTHLY ||
      def->type == TRACE_MONTHLY)
    month_dim_needed = true;
  if (def->type == SIMPLE_DAILY || def->type == PER_PFT_DAILY) 
    day_dim_needed = true;
  if (def->type == TRACE || def->type == TRACE_MONTHLY)
    trace_dim_needed = true;
  if (def->type == PER_LAYER_DAILY) {
    day_dim_needed = true;
    layer_dim_needed = true;
  }
  mean = collect.mean;
  sdev = collect.sdev;
  min = collect.min;
  max = collect.max;
  avg_buff = sdev_buff = min_buff = max_buff = 0;
  if (mean || sdev) avg_buff = new float[buff_size];
  if (sdev) sdev_buff = new float[buff_size];
  if (min) min_buff = new float[buff_size];
  if (max) max_buff = new float[buff_size];
  if (mean) mean_nc = new NcVar*[out_year_count];
  if (sdev) sdev_nc = new NcVar*[out_year_count];
  if (min) min_nc = new NcVar*[out_year_count];
  if (max) max_nc = new NcVar*[out_year_count];
  clear_buffs();
  out_index = 0;
}



LPJOutputVariable::~LPJOutputVariable()
{
  delete [] avg_buff;
  delete [] sdev_buff;
  delete [] min_buff;
  delete [] max_buff;
}


//----------------------------------------------------------------------
//
//  LPJOutputVariable::set_nc
//
//  Set up a netCDF representation of a variable in a single output
//  file.
//

void LPJOutputVariable::set_nc(int idx, NcFile &n)
{
  string sdev_nm = string(def->name) + "_SDEV";
  string min_nm = string(def->name) + "_MIN";
  string max_nm = string(def->name) + "_MAX";
  switch (def->type) {
  case SIMPLE:
    if (mean)
      mean_nc[idx] = n.add_var(def->name, ncFloat, lat_dim, lon_dim);
    if (sdev)
      sdev_nc[idx] = n.add_var(sdev_nm.c_str(), ncFloat,
                               lat_dim, lon_dim);
    if (min)
      min_nc[idx] = n.add_var(min_nm.c_str(), ncFloat,
                              lat_dim, lon_dim);
    if (max)
      max_nc[idx] = n.add_var(max_nm.c_str(), ncFloat,
                              lat_dim, lon_dim);
    break;
  case SIMPLE_MONTHLY:
    if (mean)
      mean_nc[idx] = n.add_var(def->name, ncFloat,
                               month_dim, lat_dim, lon_dim);
    if (sdev)
      sdev_nc[idx] = n.add_var(sdev_nm.c_str(), ncFloat,
                               month_dim, lat_dim, lon_dim);
    if (min)
      min_nc[idx] = n.add_var(min_nm.c_str(), ncFloat,
                              month_dim, lat_dim, lon_dim);
    if (max)
      max_nc[idx] = n.add_var(max_nm.c_str(), ncFloat,
                              month_dim, lat_dim, lon_dim);
    break;
  case SIMPLE_DAILY:
    if (mean)
      mean_nc[idx] = n.add_var(def->name, ncFloat,
                               day_dim, lat_dim, lon_dim);
    if (sdev)
      sdev_nc[idx] = n.add_var(sdev_nm.c_str(), ncFloat,
                               day_dim, lat_dim, lon_dim);
    if (min)
      min_nc[idx] = n.add_var(min_nm.c_str(), ncFloat,
                              day_dim, lat_dim, lon_dim);
    if (max)
      max_nc[idx] = n.add_var(max_nm.c_str(), ncFloat,
                              day_dim, lat_dim, lon_dim);
    break;
  case PER_PFT:
    if (mean)
      mean_nc[idx] = n.add_var(def->name, ncFloat,
                               pft_dim, lat_dim, lon_dim);
    if (sdev)
      sdev_nc[idx] = n.add_var(sdev_nm.c_str(), ncFloat,
                               pft_dim, lat_dim, lon_dim);
    if (min)
      min_nc[idx] = n.add_var(min_nm.c_str(), ncFloat,
                              pft_dim, lat_dim, lon_dim);
    if (max)
      max_nc[idx] = n.add_var(max_nm.c_str(), ncFloat,
                              pft_dim, lat_dim, lon_dim);
    break;
  case PER_PFT_MONTHLY:
    if (mean)
      mean_nc[idx] = n.add_var(def->name, ncFloat,
                               month_dim, pft_dim, lat_dim, lon_dim);
    if (sdev)
      sdev_nc[idx] = n.add_var(sdev_nm.c_str(), ncFloat,
                               month_dim, pft_dim, lat_dim, lon_dim);
    if (min)
      min_nc[idx] = n.add_var(min_nm.c_str(), ncFloat,
                              month_dim, pft_dim, lat_dim, lon_dim);
    if (max)
      max_nc[idx] = n.add_var(max_nm.c_str(), ncFloat,
                              month_dim, pft_dim, lat_dim, lon_dim);
    break;
  case PER_PFT_DAILY:
    if (mean)
      mean_nc[idx] = n.add_var(def->name, ncFloat,
                               day_dim, pft_dim, lat_dim, lon_dim);
    if (sdev)
      sdev_nc[idx] = n.add_var(sdev_nm.c_str(), ncFloat,
                               day_dim, pft_dim, lat_dim, lon_dim);
    if (min)
      min_nc[idx] = n.add_var(min_nm.c_str(), ncFloat,
                              day_dim, pft_dim, lat_dim, lon_dim);
    if (max)
      max_nc[idx] = n.add_var(max_nm.c_str(), ncFloat,
                              day_dim, pft_dim, lat_dim, lon_dim);
    break;
  case PER_LAYER_DAILY:
    if (mean)
      mean_nc[idx] = n.add_var(def->name, ncFloat,
                               day_dim, layer_dim, lat_dim, lon_dim);
    if (sdev)
      sdev_nc[idx] = n.add_var(sdev_nm.c_str(), ncFloat,
                               day_dim, layer_dim, lat_dim, lon_dim);
    if (min)
      min_nc[idx] = n.add_var(min_nm.c_str(), ncFloat,
                              day_dim, layer_dim, lat_dim, lon_dim);
    if (max)
      max_nc[idx] = n.add_var(max_nm.c_str(), ncFloat,
                              day_dim, layer_dim, lat_dim, lon_dim);
    break;
  case TRACE:
    if (mean)
      mean_nc[idx] = n.add_var(def->name, ncFloat,
                               trace_dim, lat_dim, lon_dim);
    if (sdev)
      sdev_nc[idx] = n.add_var(sdev_nm.c_str(), ncFloat,
                               trace_dim, lat_dim, lon_dim);
    if (min)
      min_nc[idx] = n.add_var(min_nm.c_str(), ncFloat,
                              trace_dim, lat_dim, lon_dim);
    if (max)
      max_nc[idx] = n.add_var(max_nm.c_str(), ncFloat,
                              trace_dim, lat_dim, lon_dim);
    break;
  case TRACE_MONTHLY:
    if (mean)
      mean_nc[idx] = n.add_var(def->name, ncFloat,
                               month_dim, trace_dim, lat_dim, lon_dim);
    if (sdev)
      sdev_nc[idx] = n.add_var(sdev_nm.c_str(), ncFloat,
                               month_dim, trace_dim, lat_dim, lon_dim);
    if (min)
      min_nc[idx] = n.add_var(min_nm.c_str(), ncFloat,
                              month_dim, trace_dim, lat_dim, lon_dim);
    if (max)
      max_nc[idx] = n.add_var(max_nm.c_str(), ncFloat,
                              month_dim, trace_dim, lat_dim, lon_dim);
    break;
  }
  if (mean && !mean_nc[idx]) {
    cout << "ERROR: Failed to add variable " << def->name << endl;
    exit(1);
  }
  if (sdev && !sdev_nc[idx]) {
    cout << "ERROR: Failed to add variable " << sdev_nm.c_str() << endl;
    exit(1);
  }
  if (min && !min_nc[idx]) {
    cout << "ERROR: Failed to add variable " << min_nm.c_str() << endl;
    exit(1);
  }
  if (max && !max_nc[idx]) {
    cout << "ERROR: Failed to add variable " << max_nm.c_str() << endl;
    exit(1);
  }

  // Changed the missing value from 1.0E20f to -9999 in an attempt to
  // make it exact which it needs to be when using it e.g. in Ferret
  if (mean) {
    mean_nc[idx]->add_att("missing_value", -9999.0f);
    mean_nc[idx]->add_att("_FillValue", -9999.0f);
  }
  if (sdev) {
    sdev_nc[idx]->add_att("missing_value", -9999.0f);
    sdev_nc[idx]->add_att("_FillValue", -9999.0f);
  }
  if (min) {
    min_nc[idx]->add_att("missing_value", -9999.0f);
    min_nc[idx]->add_att("_FillValue", -9999.0f);
  }
  if (max) {
    max_nc[idx]->add_att("missing_value", -9999.0f);
    max_nc[idx]->add_att("_FillValue", -9999.0f);
  }
}


//----------------------------------------------------------------------
//
//  Clear the averaging buffer for a variable.
//

void LPJOutputVariable::clear_buffs(void)
{
  for (int idx = 0; idx < buff_size; ++idx) {
    if (avg_buff) avg_buff[idx] = 0.0f;
    if (sdev_buff) sdev_buff[idx] = 0.0f;
    if (min_buff) min_buff[idx] = 1.0E20f;
    if (max_buff) max_buff[idx] = -1.0E20f;
  }

  accum_years = 0;
}


//----------------------------------------------------------------------
//
//  Average and output values for a variable.
//

bool LPJOutputVariable::average_and_output(void)
{
  bool retval = true;

  // Calculate mean and standard deviations as required.

  if (accum_years > 0) {
    for (int idx = 0; idx < buff_size; ++idx) {
      // At this point, sdev_buff[idx] contains the sum of squares of
      // the data and avg_buff[idx] the sum of the data.
      if (sdev) {
        sdev_buff[idx] =
          (sdev_buff[idx] * accum_years - avg_buff[idx] * avg_buff[idx]) /
          (accum_years * accum_years);
        // Protect against rounding error.
        sdev_buff[idx] = sdev_buff[idx] < 0.0 ? 0.0 : sqrt(sdev_buff[idx]);
      }
      if (mean) avg_buff[idx] /= accum_years;
    }
  }


  // Output the required data.

  switch (def->type) {
  case SIMPLE:
    if (mean)
      if (!mean_nc[out_index]->set_cur(latidx, lonidx) ||
          !mean_nc[out_index]->put(avg_buff, 1, 1))
        retval = false;
    if (sdev)
      if (!sdev_nc[out_index]->set_cur(latidx, lonidx) ||
          !sdev_nc[out_index]->put(sdev_buff, 1, 1))
        retval = false;
    if (min)
      if (!min_nc[out_index]->set_cur(latidx, lonidx) ||
          !min_nc[out_index]->put(min_buff, 1, 1))
        retval = false;
    if (max)
      if (!max_nc[out_index]->set_cur(latidx, lonidx) ||
          !max_nc[out_index]->put(max_buff, 1, 1))
        retval = false;
    break;
  case SIMPLE_MONTHLY:
    if (mean)
      if (!mean_nc[out_index]->set_cur(0, latidx, lonidx) ||
          !mean_nc[out_index]->put(avg_buff, NMON, 1, 1))
        retval = false;
    if (sdev)
      if (!sdev_nc[out_index]->set_cur(0, latidx, lonidx) ||
          !sdev_nc[out_index]->put(sdev_buff, NMON, 1, 1))
        retval = false;
    if (min)
      if (!min_nc[out_index]->set_cur(0, latidx, lonidx) ||
          !min_nc[out_index]->put(min_buff, NMON, 1, 1))
        retval = false;
    if (max)
      if (!max_nc[out_index]->set_cur(0, latidx, lonidx) ||
          !max_nc[out_index]->put(max_buff, NMON, 1, 1))
        retval = false;
    break;
  case SIMPLE_DAILY:
    if (mean)
      if (!mean_nc[out_index]->set_cur(0, latidx, lonidx) ||
          !mean_nc[out_index]->put(avg_buff, NDAYS, 1, 1))
        retval = false;
    if (sdev)
      if (!sdev_nc[out_index]->set_cur(0, latidx, lonidx) ||
          !sdev_nc[out_index]->put(sdev_buff, NDAYS, 1, 1))
        retval = false;
    if (min)
      if (!min_nc[out_index]->set_cur(0, latidx, lonidx) ||
          !min_nc[out_index]->put(min_buff, NDAYS, 1, 1))
        retval = false;
    if (max)
      if (!max_nc[out_index]->set_cur(0, latidx, lonidx) ||
          !max_nc[out_index]->put(max_buff, NDAYS, 1, 1))
        retval = false;
    break;
  case PER_PFT:
    if (mean)
      if (!mean_nc[out_index]->set_cur(0, latidx, lonidx) ||
          !mean_nc[out_index]->put(avg_buff, NPFT, 1, 1))
        retval = false;
    if (sdev)
      if (!sdev_nc[out_index]->set_cur(0, latidx, lonidx) ||
          !sdev_nc[out_index]->put(sdev_buff, NPFT, 1, 1))
        retval = false;
    if (min)
      if (!min_nc[out_index]->set_cur(0, latidx, lonidx) ||
          !min_nc[out_index]->put(min_buff, NPFT, 1, 1))
        retval = false;
    if (max)
      if (!max_nc[out_index]->set_cur(0, latidx, lonidx) ||
          !max_nc[out_index]->put(max_buff, NPFT, 1, 1))
        retval = false;
    break;
  case PER_PFT_MONTHLY:
    transpose_buffers(NMON, NPFT);
    if (mean)
      if (!mean_nc[out_index]->set_cur(0, 0, latidx, lonidx) ||
          !mean_nc[out_index]->put(avg_buff, NMON, NPFT, 1, 1))
        retval = false;
    if (sdev)
      if (!sdev_nc[out_index]->set_cur(0, 0, latidx, lonidx) ||
          !sdev_nc[out_index]->put(sdev_buff, NMON, NPFT, 1, 1))
        retval = false;
    if (min)
      if (!min_nc[out_index]->set_cur(0, 0, latidx, lonidx) ||
          !min_nc[out_index]->put(min_buff, NMON, NPFT, 1, 1))
        retval = false;
    if (max)
      if (!max_nc[out_index]->set_cur(0, 0, latidx, lonidx) ||
          !max_nc[out_index]->put(max_buff, NMON, NPFT, 1, 1))
        retval = false;
    break;
  case PER_PFT_DAILY:
    transpose_buffers(NDAYS, NPFT);
    if (mean)
      if (!mean_nc[out_index]->set_cur(0, 0, latidx, lonidx) ||
          !mean_nc[out_index]->put(avg_buff, NDAYS, NPFT, 1, 1))
        retval = false;
    if (sdev)
      if (!sdev_nc[out_index]->set_cur(0, 0, latidx, lonidx) ||
          !sdev_nc[out_index]->put(sdev_buff, NDAYS, NPFT, 1, 1))
        retval = false;
    if (min)
      if (!min_nc[out_index]->set_cur(0, 0, latidx, lonidx) ||
          !min_nc[out_index]->put(min_buff, NDAYS, NPFT, 1, 1))
        retval = false;
    if (max)
      if (!max_nc[out_index]->set_cur(0, 0, latidx, lonidx) ||
          !max_nc[out_index]->put(max_buff, NDAYS, NPFT, 1, 1))
        retval = false;
    break;
  case PER_LAYER_DAILY:
    transpose_buffers(NDAYS, NLAYER);
    if (mean)
      if (!mean_nc[out_index]->set_cur(0, 0, latidx, lonidx) ||
          !mean_nc[out_index]->put(avg_buff, NDAYS, NLAYER, 1, 1))
        retval = false;
    if (sdev)
      if (!sdev_nc[out_index]->set_cur(0, 0, latidx, lonidx) ||
          !sdev_nc[out_index]->put(sdev_buff, NDAYS, NLAYER, 1, 1))
        retval = false;
    if (min)
      if (!min_nc[out_index]->set_cur(0, 0, latidx, lonidx) ||
          !min_nc[out_index]->put(min_buff, NDAYS, NLAYER, 1, 1))
        retval = false;
    if (max)
      if (!max_nc[out_index]->set_cur(0, 0, latidx, lonidx) ||
          !max_nc[out_index]->put(max_buff, NDAYS, NLAYER, 1, 1))
        retval = false;
    break;
  case TRACE:
    if (mean)
      if (!mean_nc[out_index]->set_cur(0, latidx, lonidx) ||
          !mean_nc[out_index]->put(avg_buff, NTRACE, 1, 1))
        retval = false;
    if (sdev)
      if (!sdev_nc[out_index]->set_cur(0, latidx, lonidx) ||
          !sdev_nc[out_index]->put(sdev_buff, NTRACE, 1, 1))
        retval = false;
    if (min)
      if (!min_nc[out_index]->set_cur(0, latidx, lonidx) ||
          !min_nc[out_index]->put(min_buff, NTRACE, 1, 1))
        retval = false;
    if (max)
      if (!max_nc[out_index]->set_cur(0, latidx, lonidx) ||
          !max_nc[out_index]->put(avg_buff, NTRACE, 1, 1))
        retval = false;
    break;
  case TRACE_MONTHLY:
    transpose_buffers(NMON, NTRACE);
    if (mean)
      if (!mean_nc[out_index]->set_cur(0, 0, latidx, lonidx) ||
          !mean_nc[out_index]->put(avg_buff, NMON, NTRACE, 1, 1))
        retval = false;
    if (sdev)
      if (!sdev_nc[out_index]->set_cur(0, 0, latidx, lonidx) ||
          !sdev_nc[out_index]->put(sdev_buff, NMON, NTRACE, 1, 1))
        retval = false;
    if (min)
      if (!min_nc[out_index]->set_cur(0, 0, latidx, lonidx) ||
          !min_nc[out_index]->put(min_buff, NMON, NTRACE, 1, 1))
        retval = false;
    if (max)
      if (!max_nc[out_index]->set_cur(0, 0, latidx, lonidx) ||
          !max_nc[out_index]->put(max_buff, NMON, NTRACE, 1, 1))
        retval = false;
    break;
  }
  ++out_index;
  clear_buffs();
  return retval;
}


void LPJOutputVariable::transpose_buffers(int time_axis_len,
                                          int other_axis_len)
{
  if (mean) {
    float *tmp = new float[buff_size];
    for (int pft = 0; pft < other_axis_len; ++pft)
      for (int tim = 0; tim < time_axis_len; ++tim)
        tmp[tim * other_axis_len + pft] = avg_buff[pft * time_axis_len + tim];
    delete [] avg_buff;
    avg_buff = tmp;
  }
  if (sdev) {
    float *tmp = new float[buff_size];
    for (int pft = 0; pft < other_axis_len; ++pft)
      for (int tim = 0; tim < time_axis_len; ++tim)
        tmp[tim * other_axis_len + pft] =
          sdev_buff[pft * time_axis_len + tim];
    delete [] sdev_buff;
    sdev_buff = tmp;
  }
  if (min) {
    float *tmp = new float[buff_size];
    for (int pft = 0; pft < other_axis_len; ++pft)
      for (int tim = 0; tim < time_axis_len; ++tim)
        tmp[tim * other_axis_len + pft] = min_buff[pft * time_axis_len + tim];
    delete [] min_buff;
    min_buff = tmp;
  }
  if (max) {
    float *tmp = new float[buff_size];
    for (int pft = 0; pft < other_axis_len; ++pft)
      for (int tim = 0; tim < time_axis_len; ++tim)
        tmp[tim * other_axis_len + pft] = max_buff[pft * time_axis_len + tim];
    delete [] max_buff;
    max_buff = tmp;
  }
}


//----------------------------------------------------------------------
//
//  read_config
//
//  Read lpj.cfg configuration file and set up internal parameter
//  structure.
//

static void read_config(RunParameters &params)
{
  // Keep track of the keywords used in the file.

  set<string> keywords;


  // Open the configuration file and process a line at a time.

  FILE *fp = fopen("lpj.cfg", "r");
  if (!fp) {
    cout << "Configuration file lpj.cfg not found!" << endl;
    exit(1);
  }
  int line = 0;
  bool error = false;
  for (;;) {
    // Read lines, removing leading and trailing blanks and skipping
    // blank lines and comment lines (introduced by a # character).

    char buff[132], *pt = buff;
    fgets(buff, 132, fp);
    if (feof(fp)) break;
    ++line;
    if (strchr(buff, '\n')) *strchr(buff, '\n') = '\0';
    string in = pt;
    while (isspace(in[0])) in.erase(0, 1);
    while (isspace(in[in.size() - 1])) in.erase(in.size() - 1);
    if (in.size() == 0 || in[0] == '#') continue;


    // Split the line into the keyword and value (separated by a colon).

    string::size_type colon = in.find(':');
    if (colon == string::npos) {
      cout << "lpj.cfg:" << line
           << ": no colon separator between keyword and value" << endl;
      error = true;
      continue;
    }
    string kw = in.substr(0, colon);
    string value = in.substr(colon + 1);
    while (isspace(value[0])) value.erase(0, 1);
    while (isspace(kw[kw.size() - 1])) kw.erase(kw.size() - 1);
    for (string::size_type idx = 0; idx < kw.size(); ++idx)
      kw[idx] = toupper(kw[idx]);


    // Now split on the keyword.

    if (kw != "OUTPUT_VAR" && keywords.find(kw) != keywords.end()) {
      cout << "lpj.cfg:" << line
           << ": duplicate keyword '" << kw << "'" << endl;
      error = true;
      continue;
    }
    keywords.insert(kw);
    if (kw == "MET_DATA_TYPE") {
      if (value == "MONTHLY")
        params.met_data_type = RunParameters::MONTHLY_MET;
      else if (value == "PAR")
        params.met_data_type = RunParameters::PAR_DATA_MET;
      else if (value == "DAILY")
        params.met_data_type = RunParameters::DAILY_MET;
      else {
        cout << "lpj.cfg:" << line << ": invalid met. data type" << endl;
        cout << "value:" << value << endl;
        error = true;
      }
    }
    else if (kw == "IN_DATA_DIR") params.in_data_dir = value;
    else if (kw == "TEMP")        params.temp_file = value;
    else if (kw == "DTEMP")       params.dtemp_file = value;
    else if (kw == "PREC")        params.prec_file = value;
    else if (kw == "DPREC")       params.dprec_file = value;
    else if (kw == "SUN")         params.sun_file = value;
    else if (kw == "PAR")         params.par_file = value;
    else if (kw == "DPAR")        params.dpar_file = value;
    else if (kw == "WETDAYS")     params.wetdays_file = value;
    else if (kw == "MASK")        params.mask_file = value;
    else if (kw == "TEMP_VAR")    params.temp_var = value;
    else if (kw == "DTEMP_VAR")   params.dtemp_var = value;
    else if (kw == "PREC_VAR")    params.prec_var = value;
    else if (kw == "DPREC_VAR")   params.dprec_var = value;
    else if (kw == "SUN_VAR")     params.sun_var = value;
    else if (kw == "PAR_VAR")     params.par_var = value;
    else if (kw == "DPAR_VAR")    params.dpar_var = value;
    else if (kw == "WETDAYS_VAR") params.wetdays_var = value;
    else if (kw == "MASK_VAR")    params.mask_var = value;
    else if (kw == "SUN_IS_CLOUD")
      params.sun_is_cloud = boolean_string(value, line, error);
    else if (kw == "MASK_TYPE") {
      if (value == "boolean" || value == "BOOLEAN")
        params.mask_type = RunParameters::BOOLEAN;
      else if (value == "percent_land" || value == "PERCENT_LAND")
        params.mask_type = RunParameters::PERCENT_LAND;
      else {
        cout << "lpj.cfg:" << line << ": invalid mask type" << endl;
        error = true;
      }
    } else if (kw == "SOIL_TYPE") {
      params.soil_type_fixed = value.size() == 1 && isdigit(value[0]);
      if (params.soil_type_fixed)
        params.fixed_soil_type = value[0] - '0';
      else
        params.soil_type_file = value;
    } else if (kw == "SOIL_TYPE_VAR")  params.soil_type_var = value;
    else if (kw == "SPINUP_YEARS")
      params.spinup_years = atoi(value.c_str());
    else if (kw == "RUN_YEARS")
      params.run_years = atoi(value.c_str());
    else if (kw == "SPINUP_DATA_YEARS")
      params.spinup_data_years = atoi(value.c_str());
    else if (kw == "SPINUP_OUT_FREQ")
      params.spinup_out_freq = atoi(value.c_str());
    else if (kw == "RUN_OUT_FREQ")
      params.run_out_freq = atoi(value.c_str());
    else if (kw == "RUN_OUT_YEARS")
      params.run_out_years = atoi(value.c_str());
    else if (kw == "RUN_OUT_YEARS_ON")
      params.run_out_years_on = boolean_string(value, line, error);
    else if (kw == "CO2_CONC") {
      bool co2_fixed = true;
      for (int idx = 0; idx < value.size(); ++idx)
        if (!isdigit(value[idx]) && value[idx] != '.') {
          co2_fixed = false;
          break;
        }
      if (co2_fixed) {
        params.co2_conc.resize(1);
        params.co2_conc[0] = atof(value.c_str());
      } else
        read_co2_file(value);
    } else if (kw == "PERTURB_TMEAN") {
      params.perturb_tmean = atof(value.c_str());
    } else if (kw == "PERTURB_TRANGE") {
      params.perturb_trange = atof(value.c_str());
    } else if (kw == "PERTURB_PMEAN") {
      params.perturb_pmean = atof(value.c_str());
    } else if (kw == "OUTPUT_FILE")  params.output_file = value;
    else if (kw == "OUTPUT_VAR") {
      RunParameters::OutVariable out_var;
      string::size_type space = value.find(' ');
      if (space == string::npos) {
        // No specifiers - just collect the mean value.
        out_var.name = value;
        out_var.mean = true;
      } else {
        out_var.name = value.substr(0, space);
        value.erase(0, space);
        while (value[0] == ' ') value.erase(0, 1);
        if (value.size() == 0 ||
            value[0] != '[' || value[value.size() - 1] != ']') {
          cout << "lpj.cfg:" << line << ": invalid variable specifier" << endl;
          error = true;
        } else {
          value.erase(0, 1);
          value.erase(value.size() - 1, 1);
          vector<string> specifiers;
          while (value.find(',') != string::npos) {
            specifiers.push_back(value.substr(0, value.find(',')));
            value.erase(0, value.find(',') + 1);
          }
          specifiers.push_back(value);
          for (vector<string>::iterator it = specifiers.begin();
               it != specifiers.end(); ++it)
            if (*it == "MEAN") out_var.mean = true;
            else if (*it == "SDEV") out_var.sdev = true;
            else if (*it == "MAX") out_var.max = true;
            else if (*it == "MIN") out_var.min = true;
            else {
              cout << "lpj.cfg:" << line
                   << ": invalid variable specifier" << endl;
              error = true;
            }
        }
      }
      if (!error) params.output_vars.push_back(out_var);
    }
  }


  // Default meteorology data type.

  if (keywords.find("MET_DATA_TYPE") == keywords.end())
    params.met_data_type = RunParameters::MONTHLY_MET;


  // These entries are obligatory in the configuration file.

  const char *mand_monthly[] = { "TEMP", "PREC", "WETDAYS", "SUN", "MASK",
                                 "CO2_CONC", "OUTPUT_FILE" };
  const char *mand_par[] = { "TEMP", "PREC", "WETDAYS", "PAR", "MASK",
                             "CO2_CONC", "OUTPUT_FILE" };
  const char *mand_daily[] = { "DTEMP", "DPREC", "PAR", "TEMP", "PREC",
                               "MASK", "CO2_CONC", "OUTPUT_FILE" };
  const char **mandatory;
  int nmand;
  switch (params.met_data_type) {
  case RunParameters::MONTHLY_MET:
    mandatory = mand_monthly;
    nmand = sizeof(mand_monthly) / sizeof(char *);
    break;
  case RunParameters::PAR_DATA_MET:
    mandatory = mand_par;
    nmand = sizeof(mand_par) / sizeof(char *);
    break;
  case RunParameters::DAILY_MET:
    mandatory = mand_daily;
    nmand = sizeof(mand_daily) / sizeof(char *);
    break;
  }
  for (int check = 0; check < nmand; ++check) {
    const char *check_word = mandatory[check];
    if (keywords.find(check_word) == keywords.end()) {
      cout << "lpj.cfg: mandatory keywords '"<< check_word
           << "' missing" << endl;
      error = true;
    }
  }


  if (params.output_vars.size() != 0) {
    // Check that the specified output variables exist.

    for (vector<RunParameters::OutVariable>::iterator it =
           params.output_vars.begin();
         it != params.output_vars.end(); ++it) {
      string check_str = it->name;
      bool found = false;
      for (int chk = 0;
           chk < sizeof(LPJ_VARIABLES) / sizeof(LPJVariable); ++chk)
        if (check_str == LPJ_VARIABLES[chk].name) {
          found = true;
          break;
        }
      if (!found) {
        cout << "lpj.cfg: unrecognised variable '"<< check_str << "'" << endl;
        error = true;
      }
    }
  } else {
    // Add a default minimum set of output variables if none have been
    // specified.

    params.output_vars.push_back(RunParameters::OutVariable("fpc_grid"));
    params.output_vars.push_back(RunParameters::OutVariable("anpp"));
    params.output_vars.push_back(RunParameters::OutVariable("nind"));
    params.output_vars.push_back(RunParameters::OutVariable("gdd"));
    params.output_vars.push_back(RunParameters::OutVariable("height"));
    params.output_vars.push_back(RunParameters::OutVariable("fbare"));
  }

  // Check that any CO2 time series is of the right length.

  if (params.co2_conc.size() > 1 &&
      params.co2_conc.size() != params.run_years) {
    cout << "lpj.cfg: length of CO2 time series must match RUN_YEARS" << endl;
    error = true;
  }

  // Quit on errors.

  if (error) {
    cout << "Error in configuration file: exiting" << endl;
    exit(1);
  }
}


//----------------------------------------------------------------------
//
//  boolean_string
//
//  Convert a configuration file string value to a boolean, detecting
//  invalid values.
//

static bool boolean_string(string value, int line, bool &error_flag)
{
  if (value == "true" || value == "TRUE" ||
      value == "yes" || value == "YES" ||
      value == "t" || value == "T" ||
      value == "y" || value == "Y" ||
      value == "1")
    return true;
  else if (value == "false" || value == "FALSE" ||
      value == "no" || value == "NO" ||
      value == "f" || value == "F" ||
      value == "n" || value == "N" ||
      value == "0")
    return false;
  else {
    cout << "lpj.cfg:" << line << ": invalid boolean value" << endl;
    error_flag = true;
    return false;
  }
}


//----------------------------------------------------------------------
//
//  Grid member functions
//

Grid::Grid(string file)
{
  // Open the netCDF file.

  NcFile nc(file.c_str());
  if (!nc.is_valid()) {
    cout << "ERROR: could not open netCDF file " << file << endl;
    exit(1);
  }


  // Access latitude and longitude dimensions and variables.

  NcDim *lat_dim = nc.get_dim("lat");
  if (!lat_dim) {
    cout << "ERROR: no 'lat' dimension in netCDF file " << file << endl;
    exit(1);
  }
  NcDim *lon_dim = nc.get_dim("lon");
  if (!lon_dim) {
    cout << "ERROR: no 'lon' dimension in netCDF file " << file << endl;
    exit(1);
  }
  NcVar *lat_var = nc.get_var("lat");
  if (!lat_var) {
    cout << "ERROR: no 'lat' variable in netCDF file " << file << endl;
    exit(1);
  }
  NcVar *lon_var = nc.get_var("lon");
  if (!lon_var) {
    cout << "ERROR: no 'lon' variable in netCDF file " << file << endl;
    exit(1);
  }


  // Check that the dimensions of the coordinate variables are
  // correct.

  if (lat_var->num_dims() != 1 || lat_var->get_dim(0) != lat_dim) {
    cout << "ERROR: bad dimensions for 'lat' variable in netCDF file "
         << file << endl;
    exit(1);
  }
  if (lon_var->num_dims() != 1 || lon_var->get_dim(0) != lon_dim) {
    cout << "ERROR: bad dimensions for 'lon' variable in netCDF file "
         << file << endl;
    exit(1);
  }


  // Set up the latitude and longitude arrays.

  lats.resize(nlat = lat_dim->size());
  lons.resize(nlon = lon_dim->size());
  for (int idx = 0; idx < nlat; ++idx) lats[idx] = lat_var->as_float(idx);
  for (int idx = 0; idx < nlon; ++idx) lons[idx] = lon_var->as_float(idx);
}


Grid::Grid(const Grid &other) :
  nlat(other.nlat), nlon(other.nlon), lats(other.lats), lons(other.lons)
{}


Grid &Grid::operator=(const Grid &other)
{
  if (this != &other) {
    nlat = other.nlat;  nlon = other.nlon;
    lats = other.lats;  lons = other.lons;
  }
  return *this;
}


bool Grid::operator==(const Grid &other)
{
  return (nlat == other.nlat && nlon == other.nlon &&
          lats == other.lats && lons == other.lons);
}


//----------------------------------------------------------------------
//
//  read_mask
//
//  Read land/sea mask from netCDF file.
//

static void read_mask(BoolArray *mask, string mask_file, Grid &grid,
                      string mask_var, RunParameters::MASK_TYPE mask_type)
{
  // Open the mask file.

  NcFile nc(mask_file.c_str());
  if (!nc.is_valid()) {
    cout << "ERROR: could not open mask file " << mask_file << endl;
    exit(1);
  }


  // Access the mask variable and check the dimensions are correct.

  NcVar *var = nc.get_var(mask_var.c_str());
  NcDim *lat_dim = nc.get_dim("lat");
  NcDim *lon_dim = nc.get_dim("lon");
  if (!var) {
    cout << "ERROR: mask variable '" << mask_var
         << "' not found in mask file " << mask_file << endl;
    exit(1);
  }
  if (var->num_dims() != 2 ||
      var->get_dim(0) != lat_dim || var->get_dim(1) != lon_dim) {
    cout << "ERROR: mask variable '" << mask_var
         << "' has invalid dimensions" << endl;
    exit(1);
  }


  // Get the mask data and convert to a boolean mask.

  if (mask_type == RunParameters::BOOLEAN) {
    int tmp[grid.nlon];
    for (int latidx = 0; latidx < grid.nlat; ++latidx) {
      if (!var->set_cur(latidx, 0) || !var->get(tmp, 1, grid.nlon)) {
        cout << "ERROR: failed reading land mask" << endl;
        exit(1);
      }
      for (int lonidx = 0; lonidx < grid.nlon; ++lonidx)
        (*mask)(lonidx, latidx) = (tmp[lonidx] == 1);
    }
  } else {
    float tmp[grid.nlon];
    for (int latidx = 0; latidx < grid.nlat; ++latidx) {
      if (!var->set_cur(latidx, 0) || !var->get(tmp, 1, grid.nlon)) {
        cout << "ERROR: failed reading land mask" << endl;
        exit(1);
      }
      for (int lonidx = 0; lonidx < grid.nlon; ++lonidx)
        (*mask)(lonidx, latidx) = (tmp[lonidx] > 25.0);
    }
  }
}


//----------------------------------------------------------------------
//
//  read_soil_type
//
//  Read soil type field from netCDF file.
//

static void read_soil_type(IntArray *soil_type, string soil_type_file,
                           Grid &grid, string soil_type_var)
{
  // Open the soil type file.

  NcFile nc(soil_type_file.c_str());
  if (!nc.is_valid()) {
    cout << "ERROR: could not open soil type file " << soil_type_file << endl;
    exit(1);
  }


  // Access the soil type variable and check the dimensions are correct.

  NcVar *var = nc.get_var(soil_type_var.c_str());
  NcDim *lat_dim = nc.get_dim("lat");
  NcDim *lon_dim = nc.get_dim("lon");
  if (!var) {
    cout << "ERROR: soil type variable '" << soil_type_var
         << "' not found in soil type file " << soil_type_file << endl;
    exit(1);
  }
  if (var->num_dims() != 2 ||
      var->get_dim(0) != lat_dim || var->get_dim(1) != lon_dim) {
    cout << "ERROR: soil type variable '" << soil_type_var
         << "' has invalid dimensions" << endl;
    exit(1);
  }


  // Get the soil type data.

  for (int latidx = 0; latidx < grid.nlat; ++latidx) {
    if (!var->set_cur(latidx, 0) ||
        !var->get((*soil_type)(latidx), 1, grid.nlon)) {
      cout << "ERROR: failed reading soil type" << endl;
      exit(1);
    }
  }
}


//----------------------------------------------------------------------
//
//  read_co2_file
//
//  Read CO2 time series from an ASCII with one value per line.
//

static void read_co2_file(string fname)
{
  FILE *co2_fp = fopen(fname.c_str(), "r");
  if (!co2_fp) {
    cout << "ERROR: failed to open CO2 time series file" << endl;
    exit(1);
  }
  float co2_tmp;
  fscanf(co2_fp, "%f", &co2_tmp);
  do {
    params.co2_conc.push_back(co2_tmp);
    fscanf(co2_fp, "%f", &co2_tmp);
  } while (!feof(co2_fp));
  fclose(co2_fp);
}


//----------------------------------------------------------------------
//
//  check_var_grid
//
//  Check grid definition for a climate input variable.  Each variable
//  should have three dimensions, time, lat and lon, and the lat and
//  lon dimensions should match the overall grid.
//

static void check_var_grid(NcVar *var, NcFile *nc,
                           string var_name, string file_name)
{
  if (var->num_dims() != 3) {
    cout << "ERROR: bad dimensions for '" << var_name
         << "' variable in netCDF file " << file_name << endl;
    exit(1);
  }
  NcDim *time_dim = nc->get_dim("time");
  NcDim *lat_dim = nc->get_dim("lat");
  NcDim *lon_dim = nc->get_dim("lon");
  if (!time_dim || !lat_dim || !lon_dim) {
    cout << "ERROR: bad dimensions in netCDF file " << file_name << endl;
    exit(1);
  }
  Grid tst_grid(file_name);
  if (var->get_dim(0) != time_dim || var->get_dim(1) != lat_dim ||
      var->get_dim(2) != lon_dim || tst_grid != grid) {
    cout << "ERROR: bad dimensions for '" << var_name
         << "' variable in netCDF file " << file_name << endl;
    exit(1);
  }
}


//----------------------------------------------------------------------
//
//  open_output_file
//
//  Open and set up the output files.
//

static void open_output_files(string file_name)
{
  // Create data variable data structures.

  out_var_count = params.output_vars.size();
  out_vars = new LPJOutputVariable* [out_var_count];
  int var_idx = 0;
  for (vector<RunParameters::OutVariable>::iterator it =
         params.output_vars.begin();
       it != params.output_vars.end(); ++it) {
    string name = it->name;
    LPJVariable *def = 0;
    for (int idx = 0; idx < sizeof(LPJ_VARIABLES) / sizeof(LPJVariable);
         ++idx)
      if (name == LPJ_VARIABLES[idx].name) {
        def = &LPJ_VARIABLES[idx];
        break;
      }
    if (!def) { cerr << "INTERNAL ERROR #1" << endl;  exit(1); }
    out_vars[var_idx++] = new LPJOutputVariable(def, *it);
  }


  // Allocate space for output files.

  out_nc = new NcFile*[out_year_count];


  // Open netCDF output files.

  for (int idx = 0; idx < out_year_count; ++idx) {
    // Sort out filenames: the input string should just be a name
    // without any file extension.  Here we create an output filename
    // incorporating the index of the file.

    char index_str[32];
    sprintf(index_str, "%d", out_years[idx]);
    string out_file_name = file_name + "-" + index_str + ".nc";

    NcFile *nc = new NcFile(out_file_name.c_str(), NcFile::Replace);
    out_nc[idx] = nc;
    if (!out_nc[idx]->is_valid()) {
      cout << "ERROR: failed to open netCDF file " << out_file_name << endl;
      exit(1);
    }


    // Create dimensions.

    lat_dim = nc->add_dim("lat", grid.nlat);
    lon_dim = nc->add_dim("lon", grid.nlon);
    pft_dim = nc->add_dim("pft", NPFT);
    if (month_dim_needed) month_dim = nc->add_dim("month", NMON);
    if (day_dim_needed) day_dim = nc->add_dim("day", NDAYS);
    if (trace_dim_needed) trace_dim = nc->add_dim("trace", NTRACE);
    if (layer_dim_needed) layer_dim = nc->add_dim("layer", NLAYER);
    

    // Create coordinate variables.

    NcVar *lat_var = nc->add_var("lat", ncFloat, lat_dim);
    lat_var->add_att("long_name", "Latitude");
    lat_var->add_att("units", "degrees_north");
    NcVar *lon_var = nc->add_var("lon", ncFloat, lon_dim);
    lon_var->add_att("long_name", "Longitude");
    lon_var->add_att("units", "degrees_east");
    NcVar *pft_var = nc->add_var("pft", ncInt, pft_dim);
    pft_var->add_att("long_name", "Plant Functional Type");
    NcVar *time_var = nc->add_var("time", ncInt);
    time_var->add_att("long_name", "Time");
    time_var->add_att("units", "years");
    NcVar *month_var, *day_var, *trace_var, *layer_var;
    if (month_dim_needed) {
      month_var = nc->add_var("month", ncInt, month_dim);
      month_var->add_att("long_name", "Month of year");
      month_var->add_att("units", "months");
    }
    if (day_dim_needed) {
      day_var = nc->add_var("day", ncInt, day_dim);
      day_var->add_att("long_name", "Day of year");
      day_var->add_att("units", "days");
    }
    if (trace_dim_needed) {
      trace_var = nc->add_var("trace", ncInt, trace_dim);
      trace_var->add_att("long_name", "Trace gas index");
    }
    if (layer_dim_needed) {
      layer_var = nc->add_var("layer", ncInt, layer_dim);
      layer_var->add_att("long_name", "Layer index");
    }


    // Create data variables.

    for (int var_idx = 0; var_idx < out_var_count; ++var_idx)
      out_vars[var_idx]->set_nc(idx, *nc);


    // Write values for coordinate variables.
    
    lat_var->put(&grid.lats[0], grid.nlat);
    lon_var->put(&grid.lons[0], grid.nlon);
    int *pfts = new int[NPFT];
    for (int pft = 0; pft < NPFT; ++pft) pfts[pft] = pft + 1;
    pft_var->put(&pfts[0], NPFT);
    delete [] pfts;
    if (month_dim_needed) {
      int months[NMON] = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12 };
      month_var->put(&months[0], NMON);
    }
    if (day_dim_needed) {
      int days[NDAYS];
      for (int day = 0; day < NDAYS; ++day) days[day] = day + 1;
      day_var->put(&days[0], NDAYS);
    }
    if (trace_dim_needed) {
      int traces[NTRACE];
      for (int tr = 0; tr < NTRACE; ++tr) traces[tr] = tr + 1;
      trace_var->put(&traces[0], NTRACE);
    }
    if (layer_dim_needed) {
      int layers[NLAYER];
      for (int la = 0; la < NLAYER; ++la) layers[la] = la + 1;
      layer_var->put(&layers[0], NLAYER);
    }
    time_var->put(&out_years[idx]);

    nc->sync();
  }
}


//----------------------------------------------------------------------
//
//  handle_output_record
//
//  Handle averaging of data passed from a single call to outannual.
//

static void handle_output_record(string name, float *data)
{
  for (int idx = 0; idx < out_var_count; ++idx) {
    if (name == out_vars[idx]->def->name) {
      // Found the entry, so accumulate the data.
      for (int buf = 0; buf < out_vars[idx]->buff_size; ++buf) {
        if (out_vars[idx]->avg_buff)
          out_vars[idx]->avg_buff[buf] += data[buf];
        if (out_vars[idx]->sdev_buff)
          out_vars[idx]->sdev_buff[buf] += data[buf] * data[buf];
        if (out_vars[idx]->min_buff)
          if (data[buf] < out_vars[idx]->min_buff[buf])
            out_vars[idx]->min_buff[buf] = data[buf];
        if (out_vars[idx]->max_buff)
          if (data[buf] > out_vars[idx]->max_buff[buf])
            out_vars[idx]->max_buff[buf] = data[buf];
      }
      ++out_vars[idx]->accum_years;
    }
  }
}


//----------------------------------------------------------------------
//----------------------------------------------------------------------
//----------------------------------------------------------------------
