//----------------------------------------------------------------------
//
//  FILE:   apply-base.cpp
//  AUTHOR: I. Ross
//  DATE:   08-AUG-2005
//
//----------------------------------------------------------------------

#include <cstdio>
#include <cstdlib>
#include <netcdf.hh>
#include <iostream>

using namespace std;

#include "Array.hh"


//--------------------------------------------------------------------
//
//  CONSTANT DEFINITIONS
//

const int OUT_LON = 720, OUT_LAT = 360;


//----------------------------------------------------------------------
//
//  MAIN PROGRAM
//

int main(int argc, char *argv[])
{
  if (argc != 5) {
    cout << "Usage: apply-base <basefile> <anomfile> <var> <outfile>"
         << endl;
    exit(1);
  }
  const char *var = argv[3];
  NcFile base_nc(argv[1]);
  NcFile anom_nc(argv[2]);
  NcFile out_nc(argv[4], NcFile::Replace);
  cout << "Applying anomalies for " << var << endl;

  NcDim *in_lat_dim = anom_nc.get_dim("lat");
  NcDim *in_lon_dim = anom_nc.get_dim("lon");
  NcDim *in_time_dim = anom_nc.get_dim("time");
  NcVar *in_lat_var = anom_nc.get_var("lat");
  NcVar *in_lon_var = anom_nc.get_var("lon");
  NcVar *in_time_var = anom_nc.get_var("time");
  NcVar *base_var = base_nc.get_var(var);
  NcVar *anom_var = anom_nc.get_var(var);

  NcDim *out_lat_dim = out_nc.add_dim("lat", in_lat_dim->size());
  NcDim *out_lon_dim = out_nc.add_dim("lon", in_lon_dim->size());
  NcDim *out_time_dim = out_nc.add_dim("time");
  NcVar *out_lat_var = out_nc.add_var("lat", ncFloat, out_lat_dim);
  NcVar *out_lon_var = out_nc.add_var("lon", ncFloat, out_lon_dim);
  NcVar *out_time_var = out_nc.add_var("time", ncInt, out_time_dim);
  NcVar *out_var =
    out_nc.add_var(var, ncFloat, out_time_dim, out_lat_dim, out_lon_dim);

  out_lat_var->add_att("units", "degrees_north");
  out_lat_var->add_att("long_name", "latitude");
  out_lon_var->add_att("units", "degrees_east");
  out_lon_var->add_att("long_name", "longitude");
  out_time_var->add_att("long_name", "month");
  out_var->add_att("missing_value", -999.0f);
  out_var->add_att("_FillValue", -999.0f);

  Array3 base_data(in_lon_dim->size(), in_lat_dim->size(), 12);
  for (int mon = 0; mon < 12; ++mon)
    for (int latidx = 0; latidx < in_lat_dim->size(); ++latidx) {
      if (!base_var->set_cur(mon, latidx, 0) ||
          !base_var->get(base_data(latidx, mon), 1, 1, in_lon_dim->size())) {
        cout << "Failed reading base data" << endl;
        exit(1);
      }
    }

  float lons[in_lon_dim->size()], lats[in_lat_dim->size()];
  for (int idx = 0; idx < in_lon_dim->size(); ++idx)
    lons[idx] = in_lon_var->as_float(idx);
  for (int idx = 0; idx < in_lat_dim->size(); ++idx)
    lats[idx] = in_lat_var->as_float(idx);

  out_lat_var->put(&lats[0], in_lat_dim->size());
  out_lon_var->put(&lons[0], in_lon_dim->size());
  int times[in_time_dim->size()];
  for (int idx = 0; idx < in_time_dim->size(); ++idx) times[idx] = idx + 1;

//   printf("Year: ");
  for (int year = 1; year <= in_time_dim->size() / 12; ++year) {
//     if (year > 1) {
//       putchar('\b');
//       if (year >= 11) putchar('\b');
//       if (year >= 101) putchar('\b');
//       if (year >= 1001) putchar('\b');
//     }
//     printf("%d", year);
//     fflush(stdout);

    Array3 anom_data(in_lon_dim->size(), in_lat_dim->size(), 12);
    for (int mon = 0; mon < 12; ++mon)
      for (int latidx = 0; latidx < in_lat_dim->size(); ++latidx) {
        if (!anom_var->set_cur((year - 1) * 12 + mon, latidx, 0) ||
            !anom_var->get(anom_data(latidx, mon), 1, 1, in_lon_dim->size())) {
          cout << "Failed reading anomaly data" << endl;
          exit(1);
        }
      }

    Array3 out_data(in_lon_dim->size(), in_lat_dim->size(), 12, -999);
    for (int mon = 0; mon < 12; ++mon)
      for (int lat = 0; lat < in_lat_dim->size(); ++lat)
        for (int lon = 0; lon < in_lon_dim->size(); ++lon)
          if (base_data(lon, lat, mon) != -999 &&
              anom_data(lon, lat, mon) != -999)
            out_data(lon, lat, mon) = 
              base_data(lon, lat, mon) + anom_data(lon, lat, mon);

    for (int mon = 0; mon < 12; ++mon)
      for (int latidx = 0; latidx < in_lat_dim->size(); ++latidx) {
        if (!out_var->set_cur(12 * (year - 1) + mon, latidx, 0) ||
            !out_var->put(out_data(latidx, mon), 1, 1, in_lon_dim->size())) {
          cout << "Failed writing data" << endl;
          exit(1);
        }
      }
  }

//   putchar('\n');
  out_time_var->put(&times[0], in_time_dim->size());
  out_nc.sync();
}






