//----------------------------------------------------------------------
//
//  FILE:   regrid-motif-lpj-data.cpp
//  AUTHOR: I. Ross
//  DATE:   03-AUG-2005
//
//----------------------------------------------------------------------
//
//  Regrids netCDF file containing LPJ driver data from whatever grid
//  it's on to a 0.5 or 1.0 degree regular grid.
//
//----------------------------------------------------------------------

#include <cstdio>
#include <cstdlib>
#include <netcdf.hh>
#include <iostream>
#include <string>
using namespace std;

#include "Array.hh"


//--------------------------------------------------------------------
//
//  CONSTANT DEFINITIONS
//

int OUT_LON, OUT_LAT;


//----------------------------------------------------------------------
//
//  MAIN PROGRAM
//

int main(int argc, char *argv[])
{
  if (argc != 5) {
    cout << "Usage: regrid-motif-lpj-data <res> "
         << "<infile> <var> <outfile>" << endl;
    exit(1);
  }
  string res_str = argv[1];
  if (res_str != "0.5" && res_str != "1.0") {
    cout << "Resolution should be one of 0.5 or 1.0" << endl;
    exit(1);
  }
  float res = (res_str == "0.5") ? 0.5 : 1.0;
  const char *var = argv[3];
  NcFile in_nc(argv[2]);
  NcFile out_nc(argv[4], NcFile::Replace);
  cout << "Regridding " << argv[2] << endl;

  NcDim *in_lat_dim = in_nc.get_dim("lat");
  NcDim *in_lon_dim = in_nc.get_dim("lon");
  NcDim *in_time_dim = in_nc.get_dim("time");
  NcVar *in_lat_var = in_nc.get_var("lat");
  NcVar *in_lon_var = in_nc.get_var("lon");
  NcVar *in_time_var = in_nc.get_var("time");
  NcVar *in_var = in_nc.get_var(var);

  if (res == 0.5) { OUT_LAT = 360;  OUT_LON = 720; }
  if (res == 1.0) { OUT_LAT = 180;  OUT_LON = 360; }

  NcDim *out_lat_dim = out_nc.add_dim("lat", OUT_LAT);
  NcDim *out_lon_dim = out_nc.add_dim("lon", OUT_LON);
  NcDim *out_time_dim = out_nc.add_dim("time");
  NcVar *out_lat_var = out_nc.add_var("lat", ncFloat, out_lat_dim);
  NcVar *out_lon_var = out_nc.add_var("lon", ncFloat, out_lon_dim);
  NcVar *out_time_var = out_nc.add_var("time", ncInt, out_time_dim);
  NcVar *out_var =
    out_nc.add_var(var, ncFloat, out_time_dim, out_lat_dim, out_lon_dim);

  out_lat_var->add_att("units", "degrees_north");
  out_lat_var->add_att("long_name", "latitude");
  out_lon_var->add_att("units", "degrees_east");
  out_lon_var->add_att("long_name", "longitude");
  out_time_var->add_att("long_name", "month");

  float in_lons[in_lon_dim->size()], in_lats[in_lat_dim->size()];
  float *out_lons = new float[OUT_LON], *out_lats = new float[OUT_LAT];
  if (res == 0.5) {
    for (int idx = 0; idx < OUT_LON; ++idx)
      out_lons[idx] = 0.5 * idx + 0.25;
    for (int idx = 0; idx < OUT_LAT; ++idx)
      out_lats[idx] = 0.5 * (idx - 179.5);
  } else {
    for (int idx = 0; idx < OUT_LON; ++idx)
      out_lons[idx] = idx;
    for (int idx = 0; idx < OUT_LAT; ++idx)
      out_lats[idx] = idx - 89.75;
  }
  for (int idx = 0; idx < in_lon_dim->size(); ++idx)
    in_lons[idx] = in_lon_var->as_float(idx);
  for (int idx = 0; idx < in_lat_dim->size(); ++idx)
    in_lats[idx] = in_lat_var->as_float(idx);

  out_lat_var->put(&out_lats[0], OUT_LAT);
  out_lon_var->put(&out_lons[0], OUT_LON);
  int times[in_time_dim->size()];
  for (int idx = 0; idx < in_time_dim->size(); ++idx) times[idx] = idx + 1;

//   printf("Month: ");
  for (int month = 1; month <= in_time_dim->size(); ++month) {
//     if (month > 1) {
//       putchar('\b');
//       if (month >= 11) putchar('\b');
//       if (month >= 101) putchar('\b');
//       if (month >= 1001) putchar('\b');
//     }
//     printf("%d", month);
//     fflush(stdout);

    Array in_data(in_lon_dim->size(), in_lat_dim->size());
    for (int latidx = 0; latidx < in_lat_dim->size(); ++latidx) {
      if (!in_var->set_cur(month - 1, latidx, 0) ||
          !in_var->get(in_data(latidx), 1, 1, in_lon_dim->size())) {
        cout << "Failed reading data" << endl;
        exit(1);
      }
    }

    Array out_data(OUT_LON, OUT_LAT);
    Array::regrid_circular_x(in_data, in_lons, in_lats,
                             out_data, out_lons, out_lats, 0.0, 360.0);

    for (int latidx = 0; latidx < OUT_LAT; ++latidx) {
      if (!out_var->set_cur(month - 1, latidx, 0) ||
          !out_var->put(out_data(latidx), 1, 1, OUT_LON)) {
        cout << "Failed writing data" << endl;
        exit(1);
      }
    }
  }

//   putchar('\n');
  out_time_var->put(&times[0], in_time_dim->size());
  out_nc.sync();
}
